#!/usr/bin/env python
# -*- coding: utf-8 -*-

# pip
import os
import os.path as op
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
from matplotlib.legend import Legend
from matplotlib.patches import Circle
import matplotlib.gridspec as gridspec

# teslakit
from ..storms import GeoAzimuth

# import constants
from .config import _faspect, _fsize, _fdpi

def get_storm_linestyle(vel):
    if vel >= 60:
        return 'solid'
    elif vel >= 40:
        return 'dashed'
    elif vel >= 20:
        return 'dashdot'
    else:
        return 'dotted'

def get_storm_color(categ):

    dcs = {
        0 : 'green',
        1 : 'yellow',
        2 : 'orange',
        3 : 'red',
        4 : 'purple',
        5 : 'black',
    }

    return dcs[categ]


def axplot_Tracks_Circle(ax, lon_point, lat_point, r_point,
                         lon_storms, lat_storms, index_in, index_out,
                         categs, vel):
    'axes plot storm tracks inside radius circle'

    n_storms = len(categs)

    # plot point and  circle
    ax.plot(lon_point, lat_point, '.b', markersize=12, zorder=100)
    circle = plt.Circle(
        (lon_point, lat_point), r_point,
        edgecolor='k', fill=False)
    ax.add_patch(circle)

    # plot storms
    for lon_s, lat_s, i_i, i_o, v, c in zip(
        lon_storms, lat_storms, index_in, index_out, vel, categs):

        # get storm data
        lon_in = lon_s[i_i:i_o+1]
        lat_in = lat_s[i_i:i_o+1]

        # linestyle (velocity) and linecolor (category)
        ls = get_storm_linestyle(v)
        cs = get_storm_color(c)

        # plot storm and point at entry
        ax.plot(lon_in, lat_in, linestyle=ls, color=cs)
        ax.plot(lon_in[0], lat_in[0], '.', markersize=8, color=cs, zorder=99)

    # customize axes
    ax.axis('equal')
    ax.axis('off')
    ax.set_title(
        'Historical Tracks ({0} TCs inside R={1}º)'.format(n_storms, r_point),
        {'fontsize': 14, 'fontweight':'bold'}
    )

def axplot_Parametrized_Circle(ax, lon_point, lat_point, r_point,
                               delta_storms, gamma_storms, categs, vel):
    'axes plot storm parametrized tracks inside radius circle'

    n_storms = len(categs)

    # plot point and  circle
    ax.plot(lon_point, lat_point, '.b', markersize=12, zorder=100)
    circle = plt.Circle(
        (lon_point, lat_point), r_point,
        edgecolor='k', fill=False)
    ax.add_patch(circle)

    # TODO: def functions  track --> params / params --> track 

    # start-end points aux. opts.
    n = 1000
    ang = np.linspace(0, 2*np.pi, n)
    x_ang = r_point * np.cos(ang) + lat_point
    y_ang = r_point * np.sin(ang) + lon_point

    center_radius = np.asarray(
        [GeoAzimuth(lat_point, lon_point, x, y) for x, y in zip(x_ang, y_ang)])

    # plot storms
    for delta, gamma, v, c in zip(delta_storms, gamma_storms, vel, categs):

        # find most similar angle
        im = np.argmin(np.absolute(center_radius - delta))
        lon_entry = y_ang[im]
        lat_entry = x_ang[im]

        ent_radius = np.asarray(
            [GeoAzimuth(x, y, x_ang[im], y_ang[im]) for x, y in zip(
                x_ang, y_ang)])

        im2 = np.argmin(np.absolute(ent_radius - gamma))
        lon_exit = y_ang[im2]
        lat_exit = x_ang[im2]

        # linestyle (velocity) and linecolor (category)
        ls = get_storm_linestyle(v)
        cs = get_storm_color(c)

        # plot storm and point at entry
        ax.plot([lon_entry, lon_exit], [lat_entry, lat_exit], linestyle=ls, color=cs)
        ax.plot(lon_entry, lat_entry, '.', markersize=8, color=cs, zorder=99)

    # customize axes 
    ax.axis('equal')
    ax.axis('off')
    ax.set_title(
        'Parametrized Tracks',
        {'fontsize': 14, 'fontweight':'bold'}
    )

def axlegend_categ(ax):
    'add custom legend (storm category) to axes'

    # category legend
    lls_cat = [
        Line2D([0], [0], color = 'black'),
        Line2D([0], [0], color = 'purple'),
        Line2D([0], [0], color = 'red'),
        Line2D([0], [0], color = 'orange'),
        Line2D([0], [0], color = 'yellow'),
        Line2D([0], [0], color = 'green'),
    ]
    leg_cat = Legend(
        ax, lls_cat, ['5','4','3','2','1','0'],
        title = 'Category', bbox_to_anchor = (1.01, 1), loc='upper left',
    )
    ax.add_artist(leg_cat)

def axlegend_vel(ax):
    'add custom legend (storm velocity) to axes'

    # velocity legend
    lls_vel = [
        Line2D([0], [0], color = 'black', ls = 'solid'),
        Line2D([0], [0], color = 'black', ls = 'dashed'),
        Line2D([0], [0], color = 'black', ls = 'dashdot'),
        Line2D([0], [0], color = 'black', ls = 'dotted'),
    ]
    leg_vel = Legend(
        ax, lls_vel, ['> 60 km/h', '> 40 km/h', '> 20 km/h', '< 20 km/h'],
        title = 'Velocity', bbox_to_anchor = (1.01, 0.7), loc= 'upper left',
    )
    ax.add_artist(leg_vel)


def Plot_TCs_TracksParams(TCs_tracks, TCs_params, nm_lon='lon', nm_lat='lat', show=True):
    'Plots storms tracks and storms parametrized'

    # get data from historical tracks
    n_storms = TCs_tracks.storm.shape[0]
    lon_storms = TCs_tracks[nm_lon].values[:]
    lat_storms = TCs_tracks[nm_lat].values[:]

    # get data from parametrized tracks
    lon_point = TCs_params.point_lon
    lat_point = TCs_params.point_lat
    r_point = TCs_params.point_r

    index_in = TCs_params.index_in.values[:]
    index_out = TCs_params.index_out.values[:]
    categs_storms = TCs_params.category.values[:]
    vel_storms = TCs_params.velocity_mean.values[:]
    gamma_storms = TCs_params.gamma.values[:]
    delta_storms = TCs_params.delta.values[:]

    # figure
    fig, axs = plt.subplots(ncols=2, figsize=(_faspect*_fsize, _fsize))

    # storm tracks
    axplot_Tracks_Circle(
        axs[0], lon_point, lat_point, r_point,
        lon_storms, lat_storms, index_in, index_out,
        categs_storms, vel_storms
    )

    # storms  parametrized
    axplot_Parametrized_Circle(
        axs[1], lon_point, lat_point, r_point,
        delta_storms, gamma_storms, categs_storms, vel_storms
    )

    # small axis lims fix
    axs[0].set_xlim(axs[1].get_xlim())
    axs[0].set_ylim(axs[1].get_ylim())

    # add custom legends
    axlegend_categ(axs[1])
    axlegend_vel(axs[1])

    # show and return figure
    if show: plt.show()
    return fig

def Plot_TCs_HistoricalTracks(xds_TCs_r1, xds_TCs_r2,
                              lon1, lon2, lat1, lat2,
                              pnt_lon, pnt_lat, r1, r2,
                              nm_lon='lon', nm_lat='lat',
                              show=True):
    'Plot Historical TCs tracks map, requires basemap module'

    try:
        from mpl_toolkits.basemap import Basemap
    except:
        print('basemap module required.')
        return

    fig, ax = plt.subplots(1, figsize=(_faspect*_fsize, _fsize))

    # setup mercator map projection.
    m = Basemap(
        llcrnrlon = lon1, llcrnrlat = lat1,
        urcrnrlon = lon2, urcrnrlat = lat2,
        resolution = 'l', projection = 'cyl',
        lat_0 = lat1, lon_0 = lon1, area_thresh = 0.01,
    )
    m.drawcoastlines()
    m.fillcontinents(color = 'silver')
    m.drawmapboundary(fill_color = 'lightcyan')
    m.drawparallels(np.arange(lat1, lat2, 20), labels = [1,1,0,0])
    m.drawmeridians(np.arange(lon1, lon2, 20), labels = [0,0,0,1])

    # plot r1 storms
    for s in range(len(xds_TCs_r1.storm)):
        lon = xds_TCs_r1.isel(storm = s)[nm_lon].values[:]
        lon[np.where(lon<0)] = lon[np.where(lon<0)] + 360

        if s==0:
            ax.plot(
                lon, xds_TCs_r1.isel(storm = s)[nm_lat].values[:],
                '-', color = 'grey', alpha = 0.5,
                label = 'Enter {0}° radius'.format(r1)
            )
        else:
            ax.plot(
                lon, xds_TCs_r1.isel(storm = s)[nm_lat].values[:],
                '-', color = 'grey',alpha = 0.5
            )
            ax.plot(
                lon[0], xds_TCs_r1.isel(storm = s)[nm_lat].values[0],
                '.', color = 'grey', markersize = 10
            )

    # plot r2 storms
    for s in range(len(xds_TCs_r2.storm)):
        lon = xds_TCs_r2.isel(storm = s)[nm_lon].values[:]
        lon[np.where(lon<0)] = lon[np.where(lon<0)] + 360

        if s==0:
            ax.plot(
                lon, xds_TCs_r2.isel(storm = s)[nm_lat].values[:],
                color = 'indianred', alpha = 0.8,
                label = 'Enter {0}° radius'.format(r2)
            )
        else:
            ax.plot(
                lon, xds_TCs_r2.isel(storm = s)[nm_lat].values[:],
                color = 'indianred', alpha = 0.8
            )
            ax.plot(
                lon[0], xds_TCs_r2.isel(storm = s)[nm_lat].values[0],
                '.', color = 'indianred', markersize = 10
            )

    # plot point
    ax.plot(
        pnt_lon, pnt_lat, '.',
        markersize = 15, color = 'brown',
        label = 'STUDY SITE'
    )

    # plot r1 circle
    circle = Circle(
        m(pnt_lon, pnt_lat), r1,
        facecolor = 'grey', edgecolor = 'grey',
        linewidth = 3, alpha = 0.5,
        label='{0}° Radius'.format(r1)
    )
    ax.add_patch(circle)

    # plot r2 circle
    circle2 = Circle(
        m(pnt_lon, pnt_lat), r2,
        facecolor = 'indianred', edgecolor = 'indianred',
        linewidth = 3, alpha = 0.8,
        label='{0}° Radius'.format(r2))
    ax.add_patch(circle2)

    # customize axes
    ax.set_aspect(1.0)
    ax.set_ylim(lat1, lat2)
    ax.set_title('Historical TCs', fontsize=15)
    ax.legend(loc=0, fontsize=14)

    # show and return figure
    if show: plt.show()
    return fig

def Plot_TCs_HistoricalTracks_Category(xds_TCs_r1, cat,
                                      lon1, lon2, lat1, lat2,
                                      pnt_lon, pnt_lat, r1,
                                      nm_lon='lon', nm_lat='lat',
                                      show=True):
    'Plot Historical TCs category map, requires basemap module'

    try:
        from mpl_toolkits.basemap import Basemap
    except:
        print('basemap module required.')
        return

    fig, ax = plt.subplots(1, figsize=(_faspect*_fsize, _fsize))

    # setup mercator map projection.
    m = Basemap(
        llcrnrlon = lon1, llcrnrlat = lat1,
        urcrnrlon = lon2, urcrnrlat = lat2,
        resolution = 'l', projection = 'cyl',
        lat_0 = lat1, lon_0 = lon1, area_thresh=0.01
    )
    m.drawcoastlines()
    m.fillcontinents(color = 'silver')
    m.drawmapboundary(fill_color = 'lightcyan')
    m.drawparallels(np.arange(lat1, lat2, 20), labels = [1,0,0,0])
    m.drawmeridians(np.arange(lon1, lon2, 20), labels = [0,0,0,1])

    for s in range(len(xds_TCs_r1.storm)):
        lon = xds_TCs_r1.isel(storm = s)[nm_lon].values[:]
        lon[np.where(lon<0)] = lon[np.where(lon<0)] + 360

        if s==0:
            ax.plot(
                lon, xds_TCs_r1.isel(storm = s)[nm_lat].values[:],
                '-', color = get_storm_color(int(cat[s].values)),
                alpha = 0.5, label = 'Enter {0}° radius'.format(r1)
            )
        else:
            ax.plot(
                lon, xds_TCs_r1.isel(storm = s)[nm_lat].values[:],
                '-', color = get_storm_color(int(cat[s].values)),
                alpha = 0.5,
            )
            ax.plot(
                lon[0], xds_TCs_r1.isel(storm = s)[nm_lat].values[0],
                '.', color = get_storm_color(int(cat[s].values)),
                markersize = 10,
            )

    # plot point
    ax.plot(
        pnt_lon, pnt_lat, '.',
        markersize = 15, color = 'brown',
        label = 'STUDY SITE'
    )

    # plot circle
    circle = Circle(
        m(pnt_lon, pnt_lat), r1,
        facecolor = 'grey', edgecolor = 'grey',
        linewidth = 3, alpha = 0.5,
        label='Radius {0}º'.format(r1)
    )
    ax.add_patch(circle)

    # customize axes
    ax.set_aspect(1.0)
    ax.set_ylim(lat1,lat2)
    ax.set_title('Historical TCs', fontsize=15)
    ax.legend(loc=0, fontsize=14)
    axlegend_categ(ax)

    # show and return figure
    if show: plt.show()
    return fig

def Plot_TCs_SimTracks_Category(xds_TCs_r1, cat,
                                      lon1, lon2, lat1, lat2,
                                      pnt_lon, pnt_lat, r1,
                                      nm_lon='ylon_TC', nm_lat='ylat_TC',
                                      show=True):
    'Plot Historical TCs category map, requires basemap module'

    try:
        from mpl_toolkits.basemap import Basemap
    except:
        print('basemap module required.')
        return

    fig, ax = plt.subplots(1, figsize=[20,10])

    # setup mercator map projection.
    m = Basemap(
        llcrnrlon = lon1, llcrnrlat = lat1,
        urcrnrlon = lon2, urcrnrlat = lat2,
        resolution = 'l', projection = 'cyl',
        lat_0 = lat1, lon_0 = lon1, area_thresh=0.01
    )
    m.drawcoastlines()
    m.fillcontinents(color = 'silver')
    m.drawmapboundary(fill_color = 'lightcyan')
    m.drawparallels(np.arange(lat1, lat2, 20), labels = [1,0,0,0])
    m.drawmeridians(np.arange(lon1, lon2, 20), labels = [0,0,0,1])

    for s in range(len(xds_TCs_r1.storm)):
        lon = xds_TCs_r1.isel(storm = s)[nm_lon].values.tolist()
        lon[np.where(lon<0)] = lon[np.where(lon<0)] + 360

        if s==0:
            ax.plot(
                lon, xds_TCs_r1.isel(storm = s)[nm_lat].values.tolist(),
                '-', color = get_storm_color(int(cat[s].values)),
                alpha = 0.5, label = 'Enter {0}° radius'.format(r1)
            )
        else:
            ax.plot(
                lon, xds_TCs_r1.isel(storm = s)[nm_lat].values.tolist(),
                '-', color = get_storm_color(int(cat[s].values)),
                alpha = 0.5,
            )
            ax.plot(
                lon[0], xds_TCs_r1.isel(storm = s)[nm_lat].values.tolist()[0],
                '.', color = get_storm_color(int(cat[s].values)),
                markersize = 10,
            )

    # plot point
    ax.plot(
        pnt_lon, pnt_lat, '.',
        markersize = 15, color = 'brown',
        label = 'STUDY SITE'
    )

    # plot circle
    circle = Circle(
        m(pnt_lon, pnt_lat), r1,
        facecolor = 'grey', edgecolor = 'grey',
        linewidth = 3, alpha = 0.5,
        label='Radius {0}º'.format(r1)
    )
    ax.add_patch(circle)

    # customize axes
    ax.set_aspect(1.0)
    ax.set_ylim(lat1,lat2)
    ax.set_title('Historical TCs', fontsize=15)
    ax.legend(loc=0, fontsize=14)
    axlegend_categ(ax)

    # show and return figure
    if show: plt.show()
    return fig


def axplot_scatter_params(ax, x_hist, y_hist, x_sim, y_sim):
    'axes scatter plot variable1 vs variable2 historical and simulated'

    # simulated params 
    ax.scatter(
        x_sim, y_sim,
        c = 'silver',
        s = 3,
    )

    # historical params 
    ax.scatter(
        x_hist, y_hist,
        c = 'purple',
        s = 5,
    )

def axplot_scatter_params_MDA(ax, x_MDA, y_MDA, x_sim, y_sim):
    'axes scatter plot variable1 vs variable2 historical and simulated'

    # simulated params 
    ax.scatter(
        x_sim, y_sim,
        c = 'silver',
        s = 3,
    )

    # MDA params 
    batch_size = 100
    n = 4
    cs = ['black','red','orange','yellow','blue']

    for i in range(n):
        ix = range(i*100,(i+1)*100)

        ax.scatter(
            x_MDA[ix], y_MDA[ix],
            c = cs[i],
            s = 4,
            label = '{0}-{1}'.format(i*100,(i+1)*100-1)
        )

def axplot_histogram_params(ax, v_hist, v_sim, ttl):
    'axes histogram plot variable historical and simulated'

    # get bins
    j = np.concatenate((v_hist, v_sim))
    bins = np.linspace(np.min(j), np.max(j), 25)

    # historical
    ax.hist(
        v_hist, bins=bins,
        color = 'salmon',
        edgecolor='black',
        linewidth=1.2,
        alpha=0.5,
        label='Historical',
        density=True,
    )

    # simulated 
    ax.hist(
        v_sim, bins=bins,
        color = 'skyblue',
        edgecolor='black',
        linewidth=1.2,
        alpha=0.5,
        label='Simulated',
        density=True,
    )

    ax.set_title(ttl, fontweight='bold')
    ax.legend() #loc='upper right')


def Plot_TCs_Params_HISTvsSIM(TCs_params_hist, TCs_params_sim, show=True):
    '''
    Plot scatter with historical vs simulated parameters
    '''

    # figure conf.
    d_lab = {
        'pressure_min': 'Pmin (mbar)',
        'gamma': 'gamma (º)',
        'delta': 'delta (º)',
        'velocity_mean': 'Vmean (km/h)',
    }

    # variables to plot
    vns = ['pressure_min', 'gamma', 'delta', 'velocity_mean']
    n = len(vns)

    # figure
    fig = plt.figure(figsize=(_faspect*_fsize, _faspect*_fsize))
    gs = gridspec.GridSpec(n-1, n-1, wspace=0.2, hspace=0.2)

    for i in range(n):
        for j in range(i+1, n):

            # get variables to plot
            vn1 = vns[i]
            vn2 = vns[j]

            # historical and simulated
            vvh1 = TCs_params_hist[vn1].values[:]
            vvh2 = TCs_params_hist[vn2].values[:]

            vvs1 = TCs_params_sim[vn1].values[:]
            vvs2 = TCs_params_sim[vn2].values[:]

            # scatter plot 
            ax = plt.subplot(gs[i, j-1])
            axplot_scatter_params(ax, vvh2, vvh1, vvs2, vvs1)

            # custom labels
            if j==i+1:
                ax.set_xlabel(
                    d_lab[vn2],
                    {'fontsize':10, 'fontweight':'bold'}
                )
            if j==i+1:
                ax.set_ylabel(
                    d_lab[vn1],
                    {'fontsize':10, 'fontweight':'bold'}
                )

    # show and return figure
    if show: plt.show()
    return fig

def Plot_TCs_Params_HISTvsSIM_histogram(TCs_params_hist, TCs_params_sim,
                                        show=True):
    '''
    Plot scatter with historical vs simulated parameters
    '''

    # figure conf.
    d_lab = {
        'pressure_min': 'Pmin (mbar)',
        'gamma': 'gamma (º)',
        'delta': 'delta (º)',
        'velocity_mean': 'Vmean (km/h)',
    }

    # variables to plot
    vns = ['pressure_min', 'gamma', 'delta', 'velocity_mean']

    # figure
    fig = plt.figure(figsize=(_faspect*_fsize, _faspect*_fsize))
    gs = gridspec.GridSpec(2, 2, wspace=0.2, hspace=0.2)

    for c, vn in enumerate(vns):

        # historical and simulated
        vvh = TCs_params_hist[vn].values[:]
        vvs = TCs_params_sim[vn].values[:]

        # histograms plot 
        ax = plt.subplot(gs[c])
        axplot_histogram_params(ax, vvh, vvs, d_lab[vn])

    # show and return figure
    if show: plt.show()
    return fig

def Plot_TCs_Params_MDAvsSIM(TCs_params_MDA, TCs_params_sim, show=True):
    '''
    Plot scatter with MDA selection vs simulated parameters
    '''

    # figure conf.
    d_lab = {
        'pressure_min': 'Pmin (mbar)',
        'gamma': 'gamma (º)',
        'delta': 'delta (º)',
        'velocity_mean': 'Vmean (km/h)',
    }

    # variables to plot
    vns = ['pressure_min', 'gamma', 'delta', 'velocity_mean']
    n = len(vns)

    # figure
    fig = plt.figure(figsize=(_faspect*_fsize, _faspect*_fsize))
    gs = gridspec.GridSpec(n-1, n-1, wspace=0.2, hspace=0.2)

    for i in range(n):
        for j in range(i+1, n):

            # get variables to plot
            vn1 = vns[i]
            vn2 = vns[j]

            # historical and simulated
            vvh1 = TCs_params_MDA[vn1].values[:]
            vvh2 = TCs_params_MDA[vn2].values[:]

            vvs1 = TCs_params_sim[vn1].values[:]
            vvs2 = TCs_params_sim[vn2].values[:]

            # scatter plot 
            ax = plt.subplot(gs[i, j-1])
            axplot_scatter_params_MDA(ax, vvh2, vvh1, vvs2, vvs1)

            # custom labels
            if j==i+1:
                ax.set_xlabel(
                    d_lab[vn2],
                    {'fontsize':10, 'fontweight':'bold'}
                )
            if j==i+1:
                ax.set_ylabel(
                    d_lab[vn1],
                    {'fontsize':10, 'fontweight':'bold'}
                )

            if i==0 and j==n-1:
                ax.legend()

    # show and return figure
    if show: plt.show()
    return fig

def Plot_Category_Change(xds_categ_changeprobs, cmap='Blues', show=True):
    '''
    Plot category change betwen r1 and r2
    '''

    cp = xds_categ_changeprobs.category_change_probs.values[:]
    cs = xds_categ_changeprobs.category.values[:]

    # figure
    fig, ax = plt.subplots(1, figsize=(_faspect*_fsize/2, _faspect*_fsize/3))

    pc = ax.pcolor(cp, cmap=cmap)
    fig.colorbar(pc)

    # customize axes
    ttl = 'Category change probabilities'
    ax.set_title(ttl, {'fontsize':10, 'fontweight':'bold'})
    ax.set_xlabel('category')
    ax.set_ylabel('category')

    ticks = np.arange(cp.shape[0]) + 0.5
    ax.set_xticks(ticks)
    ax.set_yticks(ticks)
    ticks_labs = np.arange(cp.shape[0])
    ax.set_xticklabels(ticks_labs)
    ax.set_yticklabels(ticks_labs)

    # show and return figure
    if show: plt.show()
    return fig




import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import numpy as np

def SS_Distributions(xds_ss,xds_kma):
    
    clusters=np.where(np.unique(xds_kma.bmus)>=0)[0]
    n_clusters=len(clusters)
    n_rows=int(np.sqrt(n_clusters+1))
    n_cols=n_rows
    
    fig = plt.figure(figsize=[20,12])
    gs = gridspec.GridSpec(n_rows, n_cols, wspace=0.0, hspace=0.0)
    grid_row = 0
    grid_col = 0
    for ic in clusters:
        # data mean
        pos_cluster = np.where(xds_kma.bmus==ic)[0][:]
        ax = plt.subplot(gs[grid_row, grid_col])
        plt.hist(xds_ss.ss[pos_cluster],range=[-0.30, 0.30],bins=40,color='indigo',histtype='stepfilled',density=True, alpha=0.5)
        ax.set_xticks([])
        ax.set_yticks([])

        if grid_row == n_rows-1:
            ax.set_yticks([])
            ax.set_xticks(np.arange(-0.30,0.31, step=0.1))
            ax.set_xlabel('SS',fontsize=14)
        else:
            ax.set_xticks([])
            ax.set_yticks([]) 

        grid_row += 1
        if grid_row >= n_rows:
            grid_row = 0
            grid_col += 1
            
    fig = plt.figure(figsize=[20,12])
    gs = gridspec.GridSpec(n_rows, n_cols, wspace=0.0, hspace=0.0)
    grid_row = 0
    grid_col = 0
    for ic in clusters:
        # data mean
        pos_cluster = np.where(xds_kma.bmus==ic)[0][:]
        ax = plt.subplot(gs[grid_row, grid_col])
        plt.hist(xds_ss.Dwind[pos_cluster],range=[0,360],bins=50,color='darkorange',histtype='stepfilled', alpha=0.5)
        ax.set_xticks([])
        ax.set_yticks([])

        if grid_row == n_rows-1:
            ax.set_yticks([])
            ax.set_xticks(np.arange(0,361, step=90))
            ax.set_xlabel('WDir',fontsize=14)
        else:
            ax.set_xticks([])
            ax.set_yticks([])

        grid_row += 1
        if grid_row >= n_rows:
            grid_row = 0
            grid_col += 1
    
    fig = plt.figure(figsize=[20,12])
    gs = gridspec.GridSpec(n_rows, n_cols, wspace=0.0, hspace=0.0)
    grid_row = 0
    grid_col = 0
    for ic in clusters:
        # data mean
        pos_cluster = np.where(xds_kma.bmus==ic)[0][:]
        ax = plt.subplot(gs[grid_row, grid_col])
        plt.hist(xds_ss.wind[pos_cluster],range=[2,20],bins=50,color='peru',histtype='stepfilled', alpha=0.5)
        if grid_row == n_rows-1:
            ax.set_yticks([])
            ax.set_xticks(np.arange(5,21, step=5))
            ax.set_xlabel('WSpeed',fontsize=14)
        else:
            ax.set_xticks([])
            ax.set_yticks([])

        grid_row += 1
        if grid_row >= n_rows:
            grid_row = 0
            grid_col += 1


import gzip
import io
import json
import os
import shutil
from datetime import datetime
from pathlib import Path
from typing import List, Optional, Tuple, Union

import pandas as pd
import requests
import xarray as xr

from .._base_downloaders import BaseDownloader


class NOAADownloader(BaseDownloader):
    """
    This is the main class to download and read data from NOAA.

    Attributes
    ----------
    config : dict
        The configuration for NOAA data sources loaded from JSON file.
    base_path_to_download : Path
        Base path where the data is stored.
    debug : bool
        Whether to run in debug mode.

    Examples
    --------
    .. jupyter-execute::

        from bluemath_tk.downloaders.noaa.noaa_downloader import NOAADownloader

        noaa_downloader = NOAADownloader(
            base_path_to_download="/path/to/NOAA/",  # Will be created if not available
            debug=True,
            check=False,
        )

        # Download buoy bulk parameters and load DataFrame
        result = noaa_downloader.download_data(
            data_type="bulk_parameters",
            buoy_id="41001",
            years=[2020, 2021, 2022],
            load_df=True
        )
        print(result)
    """

    config = json.load(
        open(os.path.join(os.path.dirname(__file__), "NOAA_config.json"))
    )

    def __init__(
        self,
        base_path_to_download: str,
        debug: bool = True,
        check: bool = False,
    ) -> None:
        """
        Initialize the NOAA downloader.

        Parameters
        ----------
        base_path_to_download : str
            The base path to download the data to.
        debug : bool, optional
            Whether to run in debug mode. Default is True.
        check : bool, optional
            Whether to just check the data. Default is False.
        """

        super().__init__(
            base_path_to_download=base_path_to_download, debug=debug, check=check
        )
        self.set_logger_name("NOAADownloader", level="DEBUG" if debug else "INFO")

        if not self.check:
            self.logger.info("---- DOWNLOADING NOAA DATA ----")
        else:
            self.logger.info("---- CHECKING NOAA DATA ----")

    @property
    def datasets(self) -> dict:
        return self.config["datasets"]

    @property
    def data_types(self) -> dict:
        return self.config["data_types"]

    def list_data_types(self) -> List[str]:
        """
        Lists the available data types.

        Returns
        -------
        List[str]
            The list of available data types.
        """

        return list(self.data_types.keys())

    def list_datasets(self) -> List[str]:
        """
        Lists the available datasets.

        Returns
        -------
        List[str]
            The list of available datasets.
        """

        return list(self.datasets.keys())

    def show_markdown_table(self) -> None:
        """
        Create a Markdown table from the configuration dictionary and print it.
        """

        # Define the table headers
        headers = ["name", "long_name", "description", "dataset"]
        header_line = "| " + " | ".join(headers) + " |"
        separator_line = (
            "| " + " | ".join(["-" * len(header) for header in headers]) + " |"
        )

        # Initialize the table with headers
        table_lines = [header_line, separator_line]

        # Add rows for each data type
        for data_type_name, data_type_info in self.data_types.items():
            name = data_type_info.get("name", "")
            long_name = data_type_info.get("long_name", "")
            description = data_type_info.get("description", "")
            dataset = data_type_info.get("dataset", "")
            row = f"| {name} | {long_name} | {description} | {dataset} |"
            table_lines.append(row)

        # Print the table
        print("\n".join(table_lines))

    def download_data(
        self, data_type: str, load_df: bool = False, **kwargs
    ) -> Union[pd.DataFrame, xr.Dataset, str]:
        """
        Downloads the data for the specified data type.

        Parameters
        ----------
        data_type : str
            The data type to download.
            - 'bulk_parameters'
            - 'wave_spectra'
            - 'directional_spectra'
            - 'wind_forecast'

        load_df : bool, optional
            Whether to load and return the DataFrame after downloading.
            Default is False.
            If True and multiple years are specified, all years will be combined
            into a single DataFrame.
        **kwargs
            Additional keyword arguments specific to each data type.

        Returns
        -------
        Union[pd.DataFrame, xr.Dataset, str]
            Downloaded data or status message.

        Raises
        ------
        ValueError
            If the data type is not supported.
        """

        if data_type not in self.data_types:
            raise ValueError(
                f"Data type {data_type} not supported. Available types: {self.list_data_types()}"
            )

        data_type_config = self.data_types[data_type]
        dataset_config = self.datasets[data_type_config["dataset"]]

        result = None
        if data_type == "bulk_parameters":
            result = self._download_bulk_parameters(
                data_type_config, dataset_config, **kwargs
            )
            if load_df:
                buoy_id = kwargs.get("buoy_id")
                years = kwargs.get("years", [])
                if years:
                    result = self.read_bulk_parameters(buoy_id, years)
        elif data_type == "wave_spectra":
            result = self._download_wave_spectra(
                data_type_config, dataset_config, **kwargs
            )
            if load_df:
                buoy_id = kwargs.get("buoy_id")
                years = kwargs.get("years", [])
                if years:
                    result = self.read_wave_spectra(buoy_id, years)
        elif data_type == "directional_spectra":
            result = self._download_directional_spectra(
                data_type_config, dataset_config, **kwargs
            )
            if load_df:
                buoy_id = kwargs.get("buoy_id")
                years = kwargs.get("years", [])
                if years:
                    result = self.read_directional_spectra(buoy_id, years)
        elif data_type == "wind_forecast":
            result = self._download_wind_forecast(
                data_type_config, dataset_config, **kwargs
            )
        else:
            raise ValueError(f"Download for data type {data_type} not implemented")

        return result

    def _download_bulk_parameters(
        self,
        data_type_config: dict,
        dataset_config: dict,
        buoy_id: str,
        years: List[int],
        **kwargs,
    ) -> pd.DataFrame:
        """
        Download bulk parameters for a specific buoy and years.

        Parameters
        ----------
        data_type_config : dict
            The configuration for the data type.
        dataset_config : dict
            The configuration for the dataset.
        buoy_id : str
            The buoy ID.
        years : List[int]
            The years to download data for.

        Returns
        -------
        pd.DataFrame
            The downloaded data.
        """

        self.logger.info(
            f"Downloading bulk parameters for buoy {buoy_id}, years {years}"
        )

        all_data = []
        base_url = dataset_config["base_url"]

        for year in years:
            # Try main URL first, then fallbacks
            urls = [
                f"{base_url}/{data_type_config['url_pattern'].format(buoy_id=buoy_id, year=year)}"
            ]
            for fallback in data_type_config.get("fallback_urls", []):
                urls.append(f"{base_url}/{fallback.format(buoy_id=buoy_id, year=year)}")

            df = self._download_single_year_bulk(
                urls, data_type_config["columns"], year
            )
            if df is not None:
                all_data.append(df)
                self.logger.info(f"Buoy {buoy_id}: Data found for year {year}")
            else:
                self.logger.warning(
                    f"Buoy {buoy_id}: No data available for year {year}"
                )

        if all_data:
            # Combine all years
            combined_df = pd.concat(all_data, ignore_index=True)
            combined_df = combined_df.sort_values(["YYYY", "MM", "DD", "hh"])

            # Save to CSV if not in check mode
            if not self.check:
                buoy_dir = os.path.join(
                    self.base_path_to_download, "buoy_data", buoy_id
                )
                os.makedirs(buoy_dir, exist_ok=True)
                output_file = os.path.join(
                    buoy_dir, f"buoy_{buoy_id}_bulk_parameters.csv"
                )
                combined_df.to_csv(output_file, index=False)
                self.logger.info(f"Data saved to {output_file}")

                return f"Data saved to {output_file}"

            return combined_df
        else:
            self.logger.error(f"No data found for buoy {buoy_id}")
            return None

    def _download_single_year_bulk(
        self, urls: List[str], columns: List[str], year: int
    ) -> Optional[pd.DataFrame]:
        """
        Download and parse bulk parameters for a single year.

        Parameters
        ----------
        urls : List[str]
            The URLs to download the data from.
        columns : List[str]
            The columns to read from the data.
        year : int
            The year to download data for.

        Returns
        -------
        Optional[pd.DataFrame]
            The downloaded data.
        """

        for url in urls:
            try:
                response = requests.get(url)
                if response.status_code == 200:
                    content = gzip.decompress(response.content).decode("utf-8")

                    # Skip the header rows and read the data
                    data = []
                    lines = content.split("\n")[2:]  # Skip first two lines (headers)

                    # Check format by looking at the first data line
                    first_line = next(line for line in lines if line.strip())
                    cols = first_line.split()

                    # Determine format based on number of columns and year format
                    has_minutes = len(cols) == 18  # Post-2012 format has 18 columns

                    for line in lines:
                        if line.strip():
                            parts = line.split()
                            if parts:
                                # Convert 2-digit year to 4 digits if needed
                                if int(parts[0]) < 100:
                                    parts[0] = str(int(parts[0]) + 1900)

                                # Add minutes column if it doesn't exist
                                if not has_minutes:
                                    parts.insert(4, "00")

                                data.append(" ".join(parts))

                    # Read the modified data
                    df = pd.read_csv(
                        io.StringIO("\n".join(data)),
                        sep=r"\s+",
                        names=columns,
                    )

                    # Validate dates
                    valid_dates = (
                        (df["MM"] >= 1)
                        & (df["MM"] <= 12)
                        & (df["DD"] >= 1)
                        & (df["DD"] <= 31)
                        & (df["hh"] >= 0)
                        & (df["hh"] <= 23)
                        & (df["mm"] >= 0)
                        & (df["mm"] <= 59)
                    )

                    df = df[valid_dates].copy()

                    if len(df) > 0:
                        return df

            except Exception as e:
                self.logger.debug(f"Failed to download from {url}: {e}")
                continue

        return None

    def read_bulk_parameters(
        self, buoy_id: str, years: Union[int, List[int]]
    ) -> Optional[pd.DataFrame]:
        """
        Read bulk parameters for a specific buoy and year(s).

        Parameters
        ----------
        buoy_id : str
            The buoy ID.
        years : Union[int, List[int]]
            The year(s) to read data for. Can be a single year or a list of years.

        Returns
        -------
        Optional[pd.DataFrame]
            DataFrame containing the bulk parameters, or None if data not found.
        """

        if isinstance(years, int):
            years = [years]

        all_data = []
        for year in years:
            file_path = os.path.join(
                self.base_path_to_download,
                "buoy_data",
                buoy_id,
                f"buoy_{buoy_id}_bulk_parameters.csv",
            )
            try:
                df = pd.read_csv(file_path)
                df["datetime"] = pd.to_datetime(
                    df["YYYY"].astype(str)
                    + "-"
                    + df["MM"].astype(str).str.zfill(2)
                    + "-"
                    + df["DD"].astype(str).str.zfill(2)
                    + " "
                    + df["hh"].astype(str).str.zfill(2)
                    + ":"
                    + df["mm"].astype(str).str.zfill(2)
                )
                all_data.append(df)
            except FileNotFoundError:
                self.logger.error(
                    f"No bulk parameters file found for buoy {buoy_id} year {year}"
                )

        if all_data:
            return pd.concat(all_data, ignore_index=True).sort_values("datetime")
        return None

    def _download_wave_spectra(
        self,
        data_type_config: dict,
        dataset_config: dict,
        buoy_id: str,
        years: List[int],
        **kwargs,
    ) -> str:
        """
        Download wave spectra data for a specific buoy.

        Parameters
        ----------
        data_type_config : dict
            The configuration for the data type.
        dataset_config : dict
            The configuration for the dataset.
        buoy_id : str
            The buoy ID.
        years : List[int]
            The years to download data for.

        Returns
        -------
        str
            The status message.
        """

        self.logger.info(f"Downloading wave spectra for buoy {buoy_id}, years {years}")

        base_url = dataset_config["base_url"]
        buoy_dir = os.path.join(
            self.base_path_to_download, "buoy_data", buoy_id, "wave_spectra"
        )

        if not self.check:
            os.makedirs(buoy_dir, exist_ok=True)

        downloaded_files = []

        for year in years:
            url = f"{base_url}/{data_type_config['url_pattern'].format(buoy_id=buoy_id, year=year)}"

            try:
                # Read the data
                df = pd.read_csv(
                    url,
                    compression="gzip",
                    sep=r"\s+",
                    na_values=["MM", "99.00", "999.0"],
                )

                # Skip if empty or invalid data
                if df.empty or len(df.columns) < 5:
                    self.logger.warning(f"No valid data for {buoy_id} - {year}")
                    continue

                # Process datetime (simplified version)
                if not self.check:
                    output_file = os.path.join(
                        buoy_dir, f"buoy_{buoy_id}_spectra_{year}.csv"
                    )
                    df.to_csv(output_file, index=False)
                    downloaded_files.append(output_file)
                    self.logger.info(f"Successfully saved data for {buoy_id} - {year}")

            except Exception as e:
                self.logger.warning(f"No data found for: {buoy_id} - {year}: {e}")
                continue

        return f"Downloaded {len(downloaded_files)} files for wave spectra"

    def read_wave_spectra(
        self, buoy_id: str, years: Union[int, List[int]]
    ) -> Optional[pd.DataFrame]:
        """
        Read wave spectra data for a specific buoy and year(s).

        Parameters
        ----------
        buoy_id : str
            The buoy ID.
        years : Union[int, List[int]]
            The year(s) to read data for. Can be a single year or a list of years.

        Returns
        -------
        Optional[pd.DataFrame]
            DataFrame containing the wave spectra, or None if data not found
        """

        if isinstance(years, int):
            years = [years]

        all_data = []
        for year in years:
            file_path = os.path.join(
                self.base_path_to_download,
                "buoy_data",
                buoy_id,
                "wave_spectra",
                f"buoy_{buoy_id}_spectra_{year}.csv",
            )
            try:
                df = pd.read_csv(file_path)
                try:
                    df["date"] = pd.to_datetime(
                        df[["YYYY", "MM", "DD", "hh"]].rename(
                            columns={
                                "YYYY": "year",
                                "MM": "month",
                                "DD": "day",
                                "hh": "hour",
                            }
                        )
                    )
                    df.drop(columns=["YYYY", "MM", "DD", "hh"], inplace=True)
                except Exception as _e:
                    df["date"] = pd.to_datetime(
                        df[["#YY", "MM", "DD", "hh", "mm"]].rename(
                            columns={
                                "#YY": "year",
                                "MM": "month",
                                "DD": "day",
                                "hh": "hour",
                                "mm": "minute",
                            }
                        )
                    )
                    df.drop(columns=["#YY", "MM", "DD", "hh", "mm"], inplace=True)
                df.set_index("date", inplace=True)
                all_data.append(df)
            except FileNotFoundError:
                self.logger.error(
                    f"No wave spectra file found for buoy {buoy_id} year {year}"
                )

        if all_data:
            return pd.concat(all_data).sort_index()
        return None

    def _download_directional_spectra(
        self,
        data_type_config: dict,
        dataset_config: dict,
        buoy_id: str,
        years: List[int],
        **kwargs,
    ) -> str:
        """
        Download directional wave spectra coefficients.

        Parameters
        ----------
        data_type_config : dict
            The configuration for the data type.
        dataset_config : dict
            The configuration for the dataset.
        buoy_id : str
            The buoy ID.
        years : List[int]
            The years to download data for.

        Returns
        -------
        str
            The status message.
        """

        self.logger.info(
            f"Downloading directional spectra for buoy {buoy_id}, years {years}"
        )

        base_url = dataset_config["base_url"]
        coefficients = data_type_config["coefficients"]

        buoy_dir = os.path.join(
            self.base_path_to_download, "buoy_data", buoy_id, "directional_spectra"
        )
        if not self.check:
            os.makedirs(buoy_dir, exist_ok=True)

        downloaded_files = []

        for year in years:
            for coef, info in coefficients.items():
                filename = f"{buoy_id}{coef}{year}.txt.gz"
                url = f"{base_url}/{info['url_pattern'].format(buoy_id=buoy_id, year=year)}"

                if not self.check:
                    save_path = os.path.join(buoy_dir, filename)

                    try:
                        self.logger.debug(
                            f"Downloading {info['name']} data for {year}..."
                        )
                        response = requests.get(url, stream=True)
                        response.raise_for_status()

                        # Save the compressed file
                        with open(save_path, "wb") as f:
                            shutil.copyfileobj(response.raw, f)

                        downloaded_files.append(save_path)
                        self.logger.info(f"Successfully downloaded {filename}")

                    except requests.exceptions.RequestException as e:
                        self.logger.warning(f"Error downloading {filename}: {e}")
                        continue

        return f"Downloaded {len(downloaded_files)} coefficient files"

    def read_directional_spectra(
        self, buoy_id: str, years: Union[int, List[int]]
    ) -> Tuple[Optional[pd.DataFrame], ...]:
        """
        Read directional spectra data for a specific buoy and year(s).

        Parameters
        ----------
        buoy_id : str
            The buoy ID
        years : Union[int, List[int]]
            The year(s) to read data for. Can be a single year or a list of years.

        Returns
        -------
        Tuple[Optional[pd.DataFrame], ...]
            Tuple containing DataFrames for alpha1, alpha2, r1, r2, and c11,
            or None for each if data not found
        """

        if isinstance(years, int):
            years = [years]

        results = {
            "alpha1": [],
            "alpha2": [],
            "r1": [],
            "r2": [],
            "c11": [],
        }

        for year in years:
            dir_path = os.path.join(
                self.base_path_to_download,
                "buoy_data",
                buoy_id,
                "directional_spectra",
            )
            files = {
                "alpha1": f"{buoy_id}d{year}.txt.gz",
                "alpha2": f"{buoy_id}i{year}.txt.gz",
                "r1": f"{buoy_id}j{year}.txt.gz",
                "r2": f"{buoy_id}k{year}.txt.gz",
                "c11": f"{buoy_id}w{year}.txt.gz",
            }

            for name, filename in files.items():
                file_path = os.path.join(dir_path, filename)
                try:
                    df = self._read_directional_file(file_path)
                    if df is not None:
                        results[name].append(df)
                except FileNotFoundError:
                    self.logger.error(
                        f"No {name} file found for buoy {buoy_id} year {year}"
                    )

        # Combine DataFrames for each coefficient if available
        final_results = {}
        for name, dfs in results.items():
            if dfs:
                final_results[name] = pd.concat(dfs).sort_index()
            else:
                final_results[name] = None

        return (
            final_results["alpha1"],
            final_results["alpha2"],
            final_results["r1"],
            final_results["r2"],
            final_results["c11"],
        )

    def _read_directional_file(self, file_path: Path) -> Optional[pd.DataFrame]:
        """
        Read a directional spectra file and return DataFrame with datetime index.

        Parameters
        ----------
        file_path : Path
            Path to the file to read

        Returns
        -------
        Optional[pd.DataFrame]
            DataFrame containing the directional spectra data, or None if data not found
        """

        self.logger.debug(f"Reading file: {file_path}")
        try:
            with gzip.open(file_path, "rt") as f:
                # Read header lines until we find the frequencies
                header_lines = []
                while True:
                    line = f.readline().strip()
                    if not line.startswith("#") and not line.startswith("YYYY"):
                        break
                    header_lines.append(line)

                # Parse frequencies
                header = " ".join(header_lines)
                try:
                    freqs = [float(x) for x in header.split()[5:]]
                    self.logger.debug(f"Found {len(freqs)} frequencies")
                except (ValueError, IndexError) as e:
                    self.logger.error(f"Error parsing frequencies: {e}")
                    return None

                # Read data
                data = []
                dates = []
                # Process the first line
                parts = line.strip().split()
                if len(parts) >= 5:
                    try:
                        year, month, day, hour, minute = map(int, parts[:5])
                        values = [float(x) for x in parts[5:]]
                        if len(values) == len(freqs):
                            dates.append(datetime(year, month, day, hour, minute))
                            data.append(values)
                    except (ValueError, IndexError) as e:
                        self.logger.error(f"Error parsing line: {e}")

                # Read remaining lines
                for line in f:
                    parts = line.strip().split()
                    if len(parts) >= 5:
                        try:
                            year, month, day, hour, minute = map(int, parts[:5])
                            values = [float(x) for x in parts[5:]]
                            if len(values) == len(freqs):
                                dates.append(datetime(year, month, day, hour, minute))
                                data.append(values)
                        except (ValueError, IndexError) as e:
                            self.logger.error(f"Error parsing line: {e}")
                            continue

                if not data:
                    self.logger.warning("No valid data points found in file")
                    return None

                df = pd.DataFrame(data, index=dates, columns=freqs)
                self.logger.debug(f"Created DataFrame with shape: {df.shape}")
                return df

        except Exception as e:
            self.logger.error(f"Error reading file {file_path}: {str(e)}")
            return None

    def _download_wind_forecast(
        self,
        data_type_config: dict,
        dataset_config: dict,
        date: str = None,
        region: List[float] = None,
        **kwargs,
    ) -> xr.Dataset:
        """
        Download NOAA GFS wind forecast data.

        Parameters
        ----------
        data_type_config : dict
            The configuration for the data type.
        dataset_config : dict
            The configuration for the dataset.
        date : str, optional
            The date to download data for.

        Returns
        -------
        xr.Dataset
            The downloaded data.

        Notes
        -----
        - This will be DEPRECATED in the future.
        """

        if date is None:
            date = datetime.today().strftime("%Y%m%d")

        self.logger.info(f"Downloading wind forecast for date {date}")

        url_base = dataset_config["base_url"]
        dbn = "gfs_0p25_1hr"
        url = f"{url_base}/gfs{date}/{dbn}_00z"

        # File path for local storage
        forecast_dir = os.path.join(self.base_path_to_download, "wind_forecast")
        if not self.check:
            os.makedirs(forecast_dir, exist_ok=True)

        file_path = os.path.join(
            forecast_dir, f"{date}_{'_'.join(map(str, region))}.nc"
        )

        # Check if file exists
        if os.path.isfile(file_path):
            self.logger.info(
                f"File already exists: {file_path}. Loading from local storage."
            )
            data = xr.open_dataset(file_path)
        else:
            if self.check:
                self.logger.info(f"File would be downloaded to: {file_path}")
                return None

            self.logger.info(f"Downloading and cropping forecast data from: {url}")
            # Crop dataset
            data = xr.open_dataset(url)

            # Select only wind data
            variables = data_type_config["variables"]
            data_select = data[variables]

            self.logger.info(f"Storing local copy at: {file_path}")
            data_select.to_netcdf(file_path)
            data = data_select

        # Create output dataset with renamed variables
        output_vars = data_type_config["output_variables"]
        wind_data_forecast = xr.Dataset(
            {
                output_vars["u10"]: (
                    ("time", "lat", "lon"),
                    data[data_type_config["variables"][0]].values,
                ),
                output_vars["v10"]: (
                    ("time", "lat", "lon"),
                    data[data_type_config["variables"][1]].values,
                ),
            },
            coords={
                "time": data.time.values,
                "lat": data.lat.values,
                "lon": data.lon.values,
            },
        )
        wind_data_forecast["time"] = wind_data_forecast.time.dt.round("min")

        return wind_data_forecast



"""
Project: BlueMath_tk
Sub-Module: wrappers
Author: GeoOcean Research Group, Universidad de Cantabria
Repository: https://github.com/GeoOcean/BlueMath_tk.git
Status: Under development (Working)
"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
import xarray as xr

import matplotlib.pyplot as plt

from .common import GetBestRowsCols, calc_quiver

# import constants
from .config import _faspect, _fsize, _fdpi


def axplot_var_map(ax, XX, YY, vv, vd,
                   quiver=True, np_shore=np.array([]),
                   vmin=None, vmax=None):
    'plot 2D map with variable data'

    # parameters
    cmap = plt.get_cmap('seismic')

    # cplot v lims
    if vmin == None: vmin = vv.min()
    if vmax == None: vmax = vv.max()


    # plot variable 2D map
    pm = ax.pcolormesh(
        XX, YY, vv,
        cmap=cmap,
        vmin=vmin, vmax=vmax,
    )

    # optional quiver
    if quiver:
        x_q, y_q, var_q, u, v = calc_quiver(XX[0,:], YY[:,0], vv, vd, size=12)
        ax.quiver(
            x_q, y_q, -u*var_q, -v*var_q,
            width=0.003,
            #scale = 0.5,
            scale_units='inches',
        )

    # optional shoreline
    if np_shore.any():
        xs = np_shore[:,0]
        ys = np_shore[:,1]
        ax.plot(
            np_shore[:,0], np_shore[:,1],
            '.', color='dimgray',
            markersize=3, label=''
        )

        # fix axes
        ax.set_xlim(XX[0,0], XX[0,-1])
        ax.set_ylim(YY[0,0], YY[-1,0])

    # return last pcolormesh
    return pm

def scatter_maps(xds_out, var_list=[], n_cases=None, quiver=True, var_limits={},
                 np_shore=np.array([])):
    '''
    scatter plots stationary SWAN execution output for first "n_cases"

    xds_out    - swan stationary output (xarray.Dataset)

    opt. args
    var_list   - swan output variables ['Hsig', 'Tm02', 'Tpsmoo'] (default all vars)
    n_cases    - number of cases to plot (default all cases)
    quiver     - True for adding directional quiver plot
    var_limits  - dictionary with variable names as keys and a (min, max) tuple for limits
    np_shore   - shoreline, np.array x = np_shore[:,0] y = np.shore[:,1]
    '''

    # TODO improve xticks, yticks 
    # TODO legend box with info ?
    # TODO diff colormap with diff variables

    # number of cases
    if n_cases == None:
        n_cases = len(xds_out.case.values)

    # get number of rows and cols for gridplot
    n_cols, n_rows = GetBestRowsCols(n_cases)

    # allowed vars
    avs =['Hsig', 'Tm02', 'Dspr'] #, 'TPsmoo']

    # variable list 
    if var_list == []:
        var_list = dict(xds_out.variables).keys()
        var_list = [vn for vn in var_list if vn in avs]  # filter only allowed

    # mesh data to plot
    if 'lon' in xds_out.dims:
        X, Y = xds_out.lon, xds_out.lat
        xlab, ylab = 'Longitude (º)', 'Latitude (º)'
    else:
        X, Y = xds_out.X, xds_out.Y
        xlab, ylab = 'X', 'Y'
    XX, YY = np.meshgrid(X, Y)

    # iterate: one figure for each variable
    l_figs = []
    for vn in var_list:

        # plot figure
        fig, (axs) = plt.subplots(
            nrows=n_rows, ncols=n_cols,
            sharex=True, sharey=True,
            constrained_layout=False,
            figsize=(_fsize*_faspect, _fsize*_faspect),
        )

        fig.subplots_adjust(wspace=0, hspace=0)

        # common vlimits
        vmin = xds_out[vn].min()
        vmax = xds_out[vn].max()

        # optional vlimits
        if vn in var_limits.keys():
            vmin = var_limits[vn][0]
            vmax = var_limits[vn][1]

        # plot cases output
        gr, gc = 0, 0
        for ix in range(n_cases):
            out_case = xds_out.isel(case=ix)

            # variable and direction 
            vv = out_case[vn].values[:].T
            vd = out_case['Dir'].values[:].T

            # plot variable times
            ax = axs[gr, gc]
            pm = axplot_var_map(
                ax, XX, YY, vv, vd,
                quiver=quiver, np_shore=np_shore,
                vmin=vmin, vmax=vmax
            )

            # row,col counter
            gc += 1
            if gc >= n_cols:
                gc = 0
                gr += 1


        # add custom common axis labels
        fig.text(0.5, 0.04, xlab, ha='center')
        fig.text(0.04, 0.5, ylab, va='center', rotation='vertical')

        # add custom common colorbar
        cbar_ax = fig.add_axes([0.93, 0.11, 0.02, 0.77])
        fig.colorbar(pm, cax=cbar_ax)
        cbar_ax.set_ylabel(vn)


        l_figs.append(fig)

    return l_figs




import unittest

import numpy as np

from bluemath_tk.distributions._base_distributions import FitResult
from bluemath_tk.distributions.gpd import GPD


class TestGPD(unittest.TestCase):
    def setUp(self):
        self.x = np.random.rand(1000) * 2
        self.p = np.random.rand(1000)
        self.loc = 0.0
        self.scale = 1.0
        self.shape_frechet = 0.1  # GPD (Frechet)
        self.shape_weibull = -0.1  # GPD (Weibull)
        self.shape_gumbel = 0.0  # Gumbel case

    def test_pdf(self):
        for shape in [self.shape_frechet, self.shape_weibull, self.shape_gumbel]:
            custom_pdf = GPD.pdf(self.x, self.loc, self.scale, shape)
            self.assertIsInstance(custom_pdf, np.ndarray)
            self.assertEqual(custom_pdf.shape[0], 1000)

    def test_cdf(self):
        for shape in [self.shape_frechet, self.shape_weibull, self.shape_gumbel]:
            custom_cdf = GPD.cdf(self.x, self.loc, self.scale, shape)
            self.assertIsInstance(custom_cdf, np.ndarray)
            self.assertEqual(custom_cdf.shape[0], 1000)

    def test_sf(self):
        for shape in [self.shape_frechet, self.shape_weibull, self.shape_gumbel]:
            custom_sf = GPD.sf(self.x, self.loc, self.scale, shape)
            self.assertIsInstance(custom_sf, np.ndarray)
            self.assertEqual(custom_sf.shape[0], 1000)

    def test_qf(self):
        for shape in [self.shape_frechet, self.shape_weibull, self.shape_gumbel]:
            custom_qf = GPD.qf(self.p, self.loc, self.scale, shape)
            self.assertIsInstance(custom_qf, np.ndarray)
            self.assertEqual(custom_qf.shape[0], 1000)

    def test_nll(self):
        for shape in [self.shape_frechet, self.shape_weibull, self.shape_gumbel]:
            nll = GPD.nll(self.x, self.loc, self.scale, shape)
            self.assertIsInstance(nll, float)

    def test_random(self):
        for shape in [self.shape_frechet, self.shape_weibull, self.shape_gumbel]:
            random_values = GPD.random(1000, self.loc, self.scale, shape)
            self.assertIsInstance(random_values, np.ndarray)
            self.assertEqual(random_values.shape[0], 1000)

    def test_mean(self):
        for shape in [self.shape_frechet, self.shape_weibull, self.shape_gumbel]:
            mean = GPD.mean(self.loc, self.scale, shape)
            self.assertIsInstance(mean, float)

    def test_median(self):
        for shape in [self.shape_frechet, self.shape_weibull, self.shape_gumbel]:
            median = GPD.median(self.loc, self.scale, shape)
            self.assertIsInstance(median, float)

    def test_variance(self):
        for shape in [self.shape_frechet, self.shape_weibull, self.shape_gumbel]:
            variance = GPD.variance(self.loc, self.scale, shape)
            self.assertIsInstance(variance, float)

    def test_std(self):
        for shape in [self.shape_frechet, self.shape_weibull, self.shape_gumbel]:
            std = GPD.std(self.loc, self.scale, shape)
            self.assertIsInstance(std, float)

    def test_stats(self):
        for shape in [self.shape_frechet, self.shape_weibull, self.shape_gumbel]:
            stats = GPD.stats(self.loc, self.scale, shape)
            self.assertIsInstance(stats, dict)
            self.assertIn("mean", stats)
            self.assertIn("median", stats)
            self.assertIn("variance", stats)
            self.assertIn("std", stats)

    def test_invalid_scale(self):
        with self.assertRaises(ValueError):
            GPD.pdf(self.x, self.loc, 0.0, self.shape_frechet)
        with self.assertRaises(ValueError):
            GPD.cdf(self.x, self.loc, 0.0, self.shape_frechet)
        with self.assertRaises(ValueError):
            GPD.sf(self.x, self.loc, 0.0, self.shape_frechet)
        with self.assertRaises(ValueError):
            GPD.qf(self.p, self.loc, 0.0, self.shape_frechet)
        with self.assertRaises(ValueError):
            GPD.random(1000, self.loc, 0.0, self.shape_frechet)
        with self.assertRaises(ValueError):
            GPD.mean(self.loc, 0.0, self.shape_frechet)
        with self.assertRaises(ValueError):
            GPD.median(self.loc, 0.0, self.shape_frechet)
        with self.assertRaises(ValueError):
            GPD.variance(self.loc, 0.0, self.shape_frechet)
        with self.assertRaises(ValueError):
            GPD.std(self.loc, 0.0, self.shape_frechet)
        with self.assertRaises(ValueError):
            GPD.stats(self.loc, 0.0, self.shape_frechet)

    def test_fit(self):
        # Generate data using specific parameters
        loc, scale, shape = 0.5, 1.5, 0.2
        data = GPD.random(1000, loc, scale, shape, random_state=42)

        # Fit the GPD distribution to the data
        # loc is fixed at 0.0
        fit_result = GPD.fit(data - loc, f0=0.0)

        # Check the fit result
        self.assertIsInstance(fit_result, FitResult)
        self.assertTrue(fit_result.success)
        self.assertEqual(len(fit_result.params), 3)  # loc, scale, shape
        self.assertGreater(fit_result.params[1], 0)  # Scale must be > 0
        self.assertIsInstance(fit_result.nll, float)

        # Verify that the fitted parameters are close to the original ones
        self.assertAlmostEqual(fit_result.params[1], scale, delta=0.2)
        self.assertAlmostEqual(fit_result.params[2], shape, delta=0.1)


if __name__ == "__main__":
    unittest.main()



"""
Project: BlueMath_tk
Sub-Module: teslakit
Author: GeoOcean Research Group, Universidad de Cantabria
Repository: https://github.com/GeoOcean/BlueMath_tk.git
Status: OLD VERSION WORKING (MUST be updated soon)
"""

from . import (
    alr,
    climate_emulator,
    database,
    estela,
    extremes,
    intradaily,
    io,
    kma,
    mda,
    mjo,
    numerical_models,
    pca,
    plotting,
    rbf,
    statistical,
    storms,
    tides,
    util,
    waves,
)



import importlib
import logging
import os
import pickle
import sys
from abc import ABC, abstractmethod
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor, as_completed
from typing import Any, Callable, Dict, List, Tuple, Union

import numpy as np
import pandas as pd
import xarray as xr
from scipy import constants
from sklearn.metrics import (
    explained_variance_score,
    mean_absolute_error,
    mean_squared_error,
    r2_score,
)
from sklearn.preprocessing import StandardScaler

from .constants import EARTH_RADIUS
from .logging import get_file_logger
from .operations import (
    denormalize,
    destandarize,
    get_degrees_from_uv,
    get_uv_components,
    normalize,
    standarize,
)


class BlueMathModel(ABC):
    """
    Abstract base class for handling default functionalities across the project.

    This class provides core functionality used by all BlueMath models including:
    - Model saving and loading
    - Data normalization and denormalization
    - Parallel processing capabilities
    - Logging functionality
    - NaN handling
    - Directional data processing

    Attributes
    ----------
    gravity : float
        Gravitational constant from scipy.constants.
    earth_radius : float
        Earth radius in km.
    num_workers : int
        Number of parallel workers to use for processing.
    logger : logging.Logger
        Logger instance for the model.

    Notes
    -----
    All BlueMath models should inherit from this class to ensure consistent
    behavior and functionality across the project.
    """

    gravity = constants.g
    earth_radius = EARTH_RADIUS

    @abstractmethod
    def __init__(self) -> None:
        self._logger: logging.Logger = None
        self._exclude_attributes: List[str] = []
        self.num_workers: int = 1

        # [UNDER DEVELOPMENT] Below, we try to generalise parallel processing
        bluemath_num_workers = os.environ.get("BLUEMATH_NUM_WORKERS", None)
        omp_num_threads = os.environ.get("OMP_NUM_THREADS", None)
        if bluemath_num_workers is not None:
            self.logger.info(
                f"Setting self.num_workers to {bluemath_num_workers} due to BLUEMATH_NUM_WORKERS. \n"
                "Change it using self.set_num_processors_to_use method. \n"
                "Also setting OMP_NUM_THREADS to 1, to avoid conflicts with BlueMath parallel processing."
            )
            self.set_num_processors_to_use(num_processors=int(bluemath_num_workers))
            self.set_omp_num_threads(num_threads=1)
        elif omp_num_threads is not None:
            self.logger.info(
                f"Changing variable OMP_NUM_THREADS from {omp_num_threads} to 1. \n"
                f"And setting self.num_workers to {omp_num_threads}. \n"
                "To avoid conflicts with BlueMath parallel processing."
            )
            self.set_omp_num_threads(num_threads=1)
            self.set_num_processors_to_use(num_processors=int(omp_num_threads))
        else:
            self.num_workers = 1  # self.get_num_processors_available()
            self.logger.info(
                f"Setting self.num_workers to {self.num_workers}. "
                "Change it using self.set_num_processors_to_use method."
            )

    def __getstate__(self):
        """
        Control which attributes are pickled when saving the model.

        This method is automatically called by pickle.dump() to determine what
        to serialize. It excludes specified attributes and warns about xarray objects.

        Returns
        -------
        dict
            A copy of the instance's __dict__ with excluded attributes removed.

        Notes
        -----
        - Controlled by self._exclude_attributes list
        - Warns when encountering xarray objects (Dataset/DataArray)
        - Creates a deep copy of state to avoid modifying original

        See Also
        --------
        save_model : High-level method for saving model to file
        """

        state = self.__dict__.copy()
        for attr in self._exclude_attributes:
            if attr in state:
                del state[attr]
        # Iterate through the state attributes, warning about xr.Datasets
        for key, value in state.items():
            if isinstance(value, xr.Dataset) or isinstance(value, xr.DataArray):
                self.logger.warning(
                    f"Attribute {key} is an xarray Dataset / Dataarray and will be pickled!"
                )

        return state

    @property
    def logger(self) -> logging.Logger:
        """
        Get the logger instance for this model.

        Returns
        -------
        logging.Logger
            The logger instance. Creates a new file logger if none exists.

        Notes
        -----
        - Lazily instantiates logger on first access
        - Uses class name as default logger name
        - Thread-safe logger creation
        """

        if self._logger is None:
            self._logger = get_file_logger(name=self.__class__.__name__)
        return self._logger

    @logger.setter
    def logger(self, value: logging.Logger) -> None:
        """
        Set the logger instance for this model.

        Parameters
        ----------
        value : logging.Logger
            The logger instance to use.

        Raises
        ------
        ValueError
            If the logger is not an instance of logging.Logger.
        """

        if not isinstance(value, logging.Logger):
            raise ValueError("Logger must be an instance of logging.Logger")
        self._logger = value

    def set_logger_name(
        self, name: str, level: str = "INFO", console: bool = True
    ) -> None:
        """
        Configure the model's logger with a new name and settings.

        Parameters
        ----------
        name : str
            The name to give to the logger.
        level : str, optional
            The logging level to use. Default is "INFO".
            Valid values are: "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"
        console : bool, optional
            Whether to output logs to console. Default is True.

        Notes
        -----
        - Creates a new file logger with specified settings
        - Previous logger settings are overwritten
        - Log files are created in the default logging directory

        Examples
        --------
        >>> model = BlueMathModel()
        >>> model.set_logger_name("my_model", level="DEBUG", console=False)
        """

        self.logger = get_file_logger(name=name, level=level, console=console)

    def save_model(self, model_path: str, exclude_attributes: List[str] = None) -> None:
        """
        Save the model to a file using pickle.

        Parameters
        ----------
        model_path : str
            Path where the model will be saved.
        exclude_attributes : List[str], optional
            List of attribute names to exclude from saving. Default is None.
            If provided, it will override the default _exclude_attributes.

        Notes
        -----
        - Uses pickle for serialization
        - Warns if any xarray Datasets/DataArrays are being pickled
        - Creates parent directories if they don't exist
        - Excludes specified attributes from serialization

        Warnings
        --------
        - Pickle files can be security risks if loaded from untrusted sources
        - xarray objects in the model will be pickled and may be large

        Examples
        --------
        >>> model = MyBlueMathModel()
        >>> model.save_model('model.pkl', exclude_attributes=['_logger'])
        """

        self.logger.info(f"Saving model to {model_path}")
        if exclude_attributes is not None:
            self._exclude_attributes = exclude_attributes
        with open(model_path, "wb") as f:
            pickle.dump(self, f)

    def load_model(self, model_path: str) -> "BlueMathModel":
        """Loads the model from a file."""

        raise NotImplementedError(
            "This method is deprecated. Use load_model() from bluemath_tk.core.io instead."
        )

    def list_class_attributes(self) -> list:
        """
        List all non-callable attributes of the class.

        Returns
        -------
        list
            Names of all non-callable, non-private attributes.

        Notes
        -----
        - Excludes methods and private attributes (starting with __)
        - Includes properties and class variables
        - Useful for introspection and debugging

        Examples
        --------
        >>> model = BlueMathModel()
        >>> attrs = model.list_class_attributes()
        >>> print(attrs)
        ['gravity', 'num_workers', '_logger']
        """

        return [
            attr
            for attr in dir(self)
            if not callable(getattr(self, attr)) and not attr.startswith("__")
        ]

    def list_class_methods(self) -> list:
        """
        List all callable methods of the class.

        Returns
        -------
        list
            Names of all callable, non-private methods.

        Notes
        -----
        - Excludes attributes and private methods (starting with __)
        - Includes instance methods and properties
        - Useful for introspection and debugging

        Examples
        --------
        >>> model = BlueMathModel()
        >>> methods = model.list_class_methods()
        >>> print(methods)
        ['normalize', 'denormalize', 'check_nans']
        """

        return [
            attr
            for attr in dir(self)
            if callable(getattr(self, attr)) and not attr.startswith("__")
        ]

    def check_nans(
        self,
        data: Union[np.ndarray, pd.Series, pd.DataFrame, xr.DataArray, xr.Dataset],
        replace_value: Union[float, callable] = None,
        raise_error: bool = False,
    ) -> Union[np.ndarray, pd.Series, pd.DataFrame, xr.DataArray, xr.Dataset]:
        """
        Check for NaNs in the data and optionally replace them.

        Parameters
        ----------
        data : Union[np.ndarray, pd.Series, pd.DataFrame, xr.DataArray, xr.Dataset]
            The data to check for NaNs.
        replace_value : Union[float, callable], optional
            Value to replace NaNs with. If callable, the function will be called
            on the data. Default is None (no replacement).
        raise_error : bool, optional
            Whether to raise an error if NaNs are found. Default is False.

        Returns
        -------
        data : Union[np.ndarray, pd.Series, pd.DataFrame, xr.DataArray, xr.Dataset]
            The data with NaNs optionally replaced.

        Raises
        ------
        ValueError
            If NaNs are found and raise_error is True.

        Notes
        -----
        - For numpy arrays, uses np.isnan() to check for NaNs
        - For pandas objects, uses isnull() to check for NaNs
        - For xarray objects, uses isnull() to check for NaNs
        - If replace_value is callable, it takes precedence over other options

        Examples
        --------
        >>> import numpy as np
        >>> import pandas as pd
        >>> model = BlueMathModel()
        >>> df = pd.DataFrame({'a': [1, np.nan, 3]})
        >>> cleaned_df = model.check_nans(df, replace_value=0)
        >>> print(cleaned_df)
           a
        0  1
        1  0
        2  3
        """

        # If replace_value is a callable, just call and return it
        if callable(replace_value):
            self.logger.debug(f"Replace value is a callable. Calling {replace_value}.")
            return replace_value(data)
        if isinstance(data, np.ndarray):
            if np.isnan(data).any():
                if raise_error:
                    raise ValueError("Data contains NaNs.")
                self.logger.warning("Data contains NaNs.")
                if replace_value is not None:
                    data = np.nan_to_num(data, nan=replace_value)
                    self.logger.info(f"NaNs replaced with {replace_value}.")
        elif isinstance(data, pd.Series) or isinstance(data, pd.DataFrame):
            if data.isnull().values.any():
                if raise_error:
                    raise ValueError("Data contains NaNs.")
                self.logger.warning("Data contains NaNs.")
                if replace_value is not None:
                    data.fillna(replace_value, inplace=True)
                    self.logger.info(f"NaNs replaced with {replace_value}.")
        elif isinstance(data, xr.DataArray) or isinstance(data, xr.Dataset):
            if data.isnull().any():
                if raise_error:
                    raise ValueError("Data contains NaNs.")
                self.logger.warning("Data contains NaNs.")
                if replace_value is not None:
                    data = data.fillna(replace_value)
                    self.logger.info(f"NaNs replaced with {replace_value}.")
        else:
            self.logger.warning("Data type not supported for NaN check.")

        return data

    def normalize(
        self, data: Union[pd.DataFrame, xr.Dataset], custom_scale_factor: dict = {}
    ) -> Tuple[Union[pd.DataFrame, xr.Dataset], dict]:
        """
        Normalize data to 0-1 using min max scaler approach.
        More info in bluemath_tk.core.operations.normalize.

        Parameters
        ----------
        data : pd.DataFrame or xr.Dataset
            The data to normalize.
        custom_scale_factor : dict, optional
            Custom scale factors for normalization.

        Returns
        -------
        normalized_data : pd.DataFrame or xr.Dataset
            The normalized data.
        scale_factor : dict
            The scale factors used for normalization.
        """

        normalized_data, scale_factor = normalize(
            data=data, custom_scale_factor=custom_scale_factor, logger=self.logger
        )
        return normalized_data, scale_factor

    def denormalize(
        self, normalized_data: pd.DataFrame, scale_factor: dict
    ) -> pd.DataFrame:
        """
        Denormalize data using provided scale_factor.
        More info in bluemath_tk.core.operations.denormalize.

        Parameters
        ----------
        normalized_data : pd.DataFrame
            The normalized data to denormalize.
        scale_factor : dict
            The scale factors used for denormalization.

        Returns
        -------
        data : pd.DataFrame
            The denormalized data.
        """

        data = denormalize(normalized_data=normalized_data, scale_factor=scale_factor)
        return data

    def standarize(
        self,
        data: Union[np.ndarray, pd.DataFrame, xr.Dataset],
        scaler: StandardScaler = None,
        transform: bool = False,
    ) -> Tuple[Union[np.ndarray, pd.DataFrame, xr.Dataset], StandardScaler]:
        """
        Standarize data using StandardScaler.
        More info in bluemath_tk.core.operations.standarize.

        Parameters
        ----------
        data : np.ndarray, pd.DataFrame or xr.Dataset
            Input data to be standarized.
        scaler : StandardScaler, optional
            Scaler object to use for standarization. Default is None.
        transform : bool
            Whether to just transform the data. Default to False.

        Returns
        -------
        standarized_data : np.ndarray, pd.DataFrame or xr.Dataset
            Standarized data.
        scaler : StandardScaler
            Scaler object used for standarization.
        """

        standarized_data, scaler = standarize(
            data=data, scaler=scaler, transform=transform
        )
        return standarized_data, scaler

    def destandarize(
        self,
        standarized_data: Union[np.ndarray, pd.DataFrame, xr.Dataset],
        scaler: StandardScaler,
    ) -> Union[np.ndarray, pd.DataFrame, xr.Dataset]:
        """
        Destandarize data using provided scaler.
        More info in bluemath_tk.core.operations.destandarize.

        Parameters
        ----------
        standarized_data : np.ndarray, pd.DataFrame or xr.Dataset
            Standarized data to be destandarized.
        scaler : StandardScaler
            Scaler object used for standarization.

        Returns
        -------
        data : np.ndarray, pd.DataFrame or xr.Dataset
            Destandarized data.
        """

        data = destandarize(standarized_data=standarized_data, scaler=scaler)
        return data

    @staticmethod
    def get_metrics(
        data1: Union[pd.DataFrame, xr.Dataset],
        data2: Union[pd.DataFrame, xr.Dataset],
    ) -> pd.DataFrame:
        """
        Gets the metrics of the model.

        Parameters
        ----------
        data1 : pd.DataFrame or xr.Dataset
            The first dataset.
        data2 : pd.DataFrame or xr.Dataset
            The second dataset.

        Returns
        -------
        metrics : pd.DataFrame
            The metrics of the model.

        Raises
        ------
        ValueError
            If the DataFrames or Datasets have different shapes.
        TypeError
            If the inputs are not both DataFrames or both xarray Datasets.
        """

        if isinstance(data1, pd.DataFrame) and isinstance(data2, pd.DataFrame):
            if data1.shape != data2.shape:
                raise ValueError("DataFrames must have the same shape")
            variables = data1.columns
        elif isinstance(data1, xr.Dataset) and isinstance(data2, xr.Dataset):
            if sorted(list(data1.dims)) != sorted(list(data2.dims)) or sorted(
                list(data1.data_vars)
            ) != sorted(list(data2.data_vars)):
                raise ValueError(
                    "Datasets must have the same dimensions, coordinates and variables"
                )
            variables = data1.data_vars
        else:
            raise TypeError(
                "Inputs must be either both DataFrames or both xarray Datasets"
            )

        metrics = {}
        for var in variables:
            if isinstance(data1, pd.DataFrame):
                y_true = data1[var]
                y_pred = data2[var]
            else:
                y_true = data1[var].values.reshape(-1)
                y_pred = data2[var].values.reshape(-1)

            metrics[var] = {
                "mean_squared_error": mean_squared_error(y_true, y_pred),
                "r2_score": r2_score(y_true, y_pred),
                "mean_absolute_error": mean_absolute_error(y_true, y_pred),
                "explained_variance_score": explained_variance_score(y_true, y_pred),
            }

        return pd.DataFrame(metrics).T

    @staticmethod
    def get_uv_components(x_deg: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
        This method calculates the u and v components for the given directional data.

        Here, we assume that the directional data is in degrees,
            beign 0° the North direction,
            and increasing clockwise.

                   0° N
                    |
                    |
        270° W <---------> 90° E
                    |
                    |
                  90° S

        Parameters
        ----------
        x_deg : np.ndarray
            The directional data in degrees.

        Returns
        -------
        Tuple[np.ndarray, np.ndarray]
            The u and v components.
        """

        return get_uv_components(x_deg)

    @staticmethod
    def get_degrees_from_uv(xu: np.ndarray, xv: np.ndarray) -> np.ndarray:
        """
        This method calculates the degrees from the u and v components.

        Here, we assume u and v represent angles between 0 and 360 degrees,
            where 0° is the North direction,
            and increasing clockwise.

                     (u=0, v=1)
                         |
                         |
        (u=-1, v=0) <---------> (u=1, v=0)
                         |
                         |
                     (u=0, v=-1)

        Parameters
        ----------
        xu : np.ndarray
            The u component.
        xv : np.ndarray
            The v component.

        Returns
        -------
        np.ndarray
            The degrees.
        """

        return get_degrees_from_uv(xu, xv)

    def set_omp_num_threads(self, num_threads: int) -> None:
        """
        Set the number of OpenMP threads for parallel operations.

        Parameters
        ----------
        num_threads : int
            Number of OpenMP threads to use.

        Notes
        -----
        - Sets the OMP_NUM_THREADS environment variable
        - Reloads numpy to ensure new thread settings take effect
        - May affect other libraries using OpenMP

        Warnings
        --------
        - This method is under development and behavior may change
        - Reloading numpy may have side effects in running calculations

        See Also
        --------
        set_num_processors_to_use : Set number of processors for BlueMath parallel processing
        """

        os.environ["OMP_NUM_THREADS"] = str(num_threads)

        # Re-import numpy if it is already imported
        if "numpy" in sys.modules:
            importlib.reload(np)

    def get_num_processors_available(self) -> int:
        """
        Gets the number of processors available.

        Returns
        -------
        int
            The number of processors available.

        TODO
        ----
        - Check whether available processors are used or not.
        """

        return int(os.cpu_count() * 0.9)

    def set_num_processors_to_use(self, num_processors: int) -> None:
        """
        Set the number of processors to use for parallel processing.

        Parameters
        ----------
        num_processors : int
            Number of processors to use. If -1, uses all available processors
            minus one for system processes.

        Raises
        ------
        ValueError
            If num_processors is <= 0 (except -1).

        Notes
        -----
        - Automatically adjusts if requesting too many processors
        - Sets the num_workers attribute used by parallel processing methods
        - Takes into account system resources to avoid overload

        See Also
        --------
        get_num_processors_available : Get number of available processors
        parallel_execute : Execute functions in parallel
        """

        # Retrieve the number of processors available
        num_processors_available = self.get_num_processors_available()

        # Check if the number of processors requested is valid
        if num_processors == -1:
            num_processors = num_processors_available
        elif num_processors <= 0:
            raise ValueError("Number of processors must be greater than 0")
        elif (num_processors_available - num_processors) < 2:
            self.logger.info(
                "Number of processors requested leaves less than 2 processors available"
            )

        # Set the number of processors to use
        self.num_workers = num_processors

    def parallel_execute(
        self,
        func: Callable,
        items: List[Any],
        num_workers: int,
        cpu_intensive: bool = False,
        **kwargs,
    ) -> Dict[int, Any]:
        """
        Execute a function in parallel across multiple items.

        Parameters
        ----------
        func : Callable
            The function to execute. Should accept single item and **kwargs.
        items : List[Any]
            List of items to process in parallel.
        num_workers : int
            Number of parallel workers to use.
        cpu_intensive : bool, optional
            If True, uses ProcessPoolExecutor, otherwise ThreadPoolExecutor.
            Default is False.
        **kwargs : dict
            Additional keyword arguments passed to func.

        Returns
        -------
        Dict[int, Any]
            Dictionary mapping item indices to function results.

        Raises
        ------
        Exception
            Any exception raised by func is logged and the job continues.

        Notes
        -----
        - Uses ThreadPoolExecutor for I/O-bound tasks
        - Uses ProcessPoolExecutor for CPU-bound tasks
        - Results maintain original item order via index mapping
        - Failed jobs are logged but don't stop execution

        Warnings
        --------
        - ThreadPoolExecutor may have GIL limitations
        - ProcessPoolExecutor doesn't work with non-picklable objects
        - File operations may fail with ThreadPoolExecutor

        Examples
        --------
        >>> def square(x):
        ...     return x * x
        >>> model = BlueMathModel()
        >>> results = model.parallel_execute(square, [1, 2, 3], num_workers=2)
        >>> print(results)
        {0: 1, 1: 4, 2: 9}
        """

        results = {}

        executor_class = ProcessPoolExecutor if cpu_intensive else ThreadPoolExecutor
        self.logger.info(f"Using {executor_class.__name__} for parallel execution")

        with executor_class(max_workers=num_workers) as executor:
            future_to_item = {
                executor.submit(func, *item, **kwargs)
                if isinstance(item, tuple)
                else executor.submit(func, item, **kwargs): i
                for i, item in enumerate(items)
            }
            for future in as_completed(future_to_item):
                i = future_to_item[future]
                try:
                    result = future.result()
                    results[i] = result
                except Exception as exc:
                    self.logger.error(f"Job for {i} generated an exception: {exc}")

        return results



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# common
import os
import os.path as op

# pip
import numpy as np

import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from mpl_toolkits.mplot3d import Axes3D

# teslakit
from .custom_colors import GetClusterColors

# import constants
from .config import _faspect, _fsize, _fdpi


def axplot_PC_hist(ax, pc_wt, color_wt, nb=30, ylab=None):
    'axes plot AWT singular PC histogram'

    ax.hist(pc_wt, nb, density=True, color=color_wt)

    # gridlines and axis properties
    ax.grid(True, which='both', axis='both', linestyle='--', color='grey')
    ax.set_xlim([-3,3])
    ax.set_yticklabels([])
    ax.tick_params(axis='x', which='major', labelsize=5)
    if ylab:
        ax.set_ylabel(ylab, {'fontweight':'bold'}, labelpad=-3)

def axplot_PCs_2D(ax, PC1, PC2, d_wts, c_wts):
    'axes plot AWT PCs 1,2,3 (3D)'

    # calculate PC centroids
    pc1_wt = [np.mean(PC1[d_wts[i]]) for i in sorted(d_wts.keys())]
    pc2_wt = [np.mean(PC2[d_wts[i]]) for i in sorted(d_wts.keys())]

    # scatter  plot
    ax.scatter(
        PC1, PC2,
        c = 'silver',
        s = 3,
    )

    # WT centroids
    for x,y,c in zip(pc1_wt, pc2_wt, c_wts):
        ax.scatter(x, y, c=[c], s=10)

    ax.set_xlim([-3,3])
    ax.set_ylim([-3,3])
    ax.set_xticks([])
    ax.set_yticks([])

def axplot_PCs_3D(ax, pcs_wt, color_wt, ttl='PCs'):
    'axes plot AWT PCs 1,2,3 (3D)'

    PC1 = pcs_wt[:,0]
    PC2 = pcs_wt[:,1]
    PC3 = pcs_wt[:,2]

    # scatter  plot
    ax.scatter(
        PC1, PC2, PC3,
        c = [color_wt],
        s = 2,
    )

    ax.set_xlim([-3,3])
    ax.set_ylim([-3,3])
    ax.set_zlim([-3,3])
    ax.set_xticklabels([])
    ax.set_yticklabels([])
    ax.set_zticklabels([])
    ax.set_title(ttl, {'fontsize':8, 'fontweight':'bold'})

def axplot_PCs_3D_WTs(ax, d_PCs, wt_colors, ttl='PCs'):
    'axes plot AWT PCs 1,2,3 (3D)'

    # plot each weather type
    wt_keys = sorted(d_PCs.keys())
    for ic, k in enumerate(wt_keys):
        PC1 = d_PCs[k][:,0]
        PC2 = d_PCs[k][:,1]
        PC3 = d_PCs[k][:,2]

        # scatter  plot
        ax.scatter(
            PC1, PC2, PC3,
            c = [wt_colors[ic]],
            label = k,
            s = 6,
        )

    ax.set_xlabel('PC1', {'fontsize':10})
    ax.set_ylabel('PC2', {'fontsize':10})
    ax.set_zlabel('PC3', {'fontsize':10})
    
    lim = np.nanmax([np.nanmax(np.abs(PC1)), np.nanmax(np.abs(PC2)), np.nanmax(np.abs(PC3))])
    
#     ax.set_xlim([-3,3])
#     ax.set_ylim([-3,3])
#     ax.set_zlim([-3,3])

    
    ax.set_xlim([-lim, lim])
    ax.set_ylim([-lim, lim])
    ax.set_zlim([-lim, lim])
    ax.set_title(ttl, {'fontsize':10, 'fontweight':'bold'})


def Plot_PCs_WT(PCs, variance, bmus, n_clusters, n=3, show=True):
    '''
    Plot Annual Weather Types PCs using 2D axis

    PCs, variance, bmus - from KMA_simple() or KMA_regression_guided()
    n                   - number of PCs to plot
    '''

    # get cluster colors
    cs_wt = GetClusterColors(n_clusters)

    # get cluster - bmus indexes
    d_wts = {}
    for i in range(n_clusters):
        d_wts[i] = np.where(bmus == i)[:]

    # figure
    fig = plt.figure(figsize=(_faspect*_fsize, _faspect*_fsize))
    gs = gridspec.GridSpec(n-1, n-1, wspace=0.0, hspace=0.0)

    for i in range(n):
        for j in range(i+1, n):

            # get PCs to plot
            PC1 = np.divide(PCs[:,i], np.sqrt(variance[i]))
            PC2 = np.divide(PCs[:,j], np.sqrt(variance[j]))

            # plot PCs (2D)
            ax = plt.subplot(gs[i, j-1])
            axplot_PCs_2D(ax, PC1, PC2, d_wts, cs_wt)

            # custom labels
            if i==0:
                ax.set_xlabel(
                    'PC {0}'.format(j+1),
                    {'fontsize':10, 'fontweight':'bold'}
                )
                ax.xaxis.set_label_position('top')
            if j==n-1:
                ax.set_ylabel(
                    'PC {0}'.format(i+1),
                    {'fontsize':10, 'fontweight':'bold'}
                )
                ax.yaxis.set_label_position('right')

    # show and return figure
    if show: plt.show()
    return fig

def Plot_PCs_Compare_3D(d_PCs_fit, d_PCs_rnd, show=True):
    '''
    Plot Annual Weather Types PCs fit - PCs rnd comparison (3D)
    '''

    # get cluster colors
    n_clusters = len(d_PCs_fit.keys())
    cs_wt = GetClusterColors(n_clusters)

    # figure
    fig = plt.figure(figsize=(_faspect*_fsize, _fsize/1.66))
    gs = gridspec.GridSpec(1, 2, wspace=0.10, hspace=0.35)
    ax_fit = plt.subplot(gs[0, 0], projection='3d')
    ax_sim = plt.subplot(gs[0, 1], projection='3d')

    # Plot PCs (3D)
    axplot_PCs_3D_WTs(ax_fit, d_PCs_fit,  cs_wt, ttl='PCs fit')
    axplot_PCs_3D_WTs(ax_sim, d_PCs_rnd,  cs_wt, ttl='PCs sim')

    # show and return figure
    if show: plt.show()
    return fig

def Plot_WT_PCs_3D(d_PCs, n_clusters, show=True):
    '''
    Plot Weather Types PCs (3D)
    '''

    # get cluster colors
    cs_wt = GetClusterColors(n_clusters)

    # figure
    fig = plt.figure(figsize=(_faspect*_fsize, _fsize*2/1.66))
    gs = gridspec.GridSpec(1, 1, wspace=0.10, hspace=0.35)
    ax_pcs = plt.subplot(gs[0, 0], projection='3d')

    # Plot PCs (3D)
    axplot_PCs_3D_WTs(ax_pcs, d_PCs,  cs_wt, ttl='PCs')

    # show and return figure
    if show: plt.show()
    return fig



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# common
import os
import os.path as op
from datetime import datetime, timedelta

# pip
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import matplotlib.dates as mdates
# from matplotlib.colors import DivergingNorm
from scipy.interpolate import interp1d
from scipy.stats import  gumbel_l, genextreme

# teslakit
from .outputs import axplot_compare_histograms

# import constants
from .config import _faspect, _fsize, _fdpi

# TODO: REFACTOR CHROM Y SIGMA


# Climate Emulator: Fit Report

def Plot_GEVParams(xda_gev_var, c_shape='bwr', c_other='hot_r', show=True):
    'Plot GEV params for a GEV parameter variable (sea_Hs, swell_1_Hs, ...)'

    name = xda_gev_var.name
    params = xda_gev_var.parameter.values[:]
    ss = int(np.sqrt(len(xda_gev_var.n_cluster)))  # this will fail if cant sqrt

    # plot figure
    fig, axs = plt.subplots(2,2, figsize=(_faspect*_fsize, _fsize))
    axs = [i for sl in axs for i in sl]

    # empty last axis
    axs[3].axis('off')

    for c, par in enumerate(params):
        ax = axs[c]

        par_values = xda_gev_var.sel(parameter=par).values[:]
        par_values[par_values==1.0e-10]=0

        rr_pv = np.flipud(np.reshape(par_values,(ss,ss)).T)

        if par == 'shape':
            max_value = max(abs(np.min(par_values)), abs(np.max(par_values)))
            cl = [-max_value, max_value]
            if cl[0]>=0: cl[0]=-0.000000001
            if cl[1]<=0: cl[1]=+0.000000001
            # norm = DivergingNorm(vmin=cl[0], vcenter=0, vmax=cl[1])
            norm = None
            cma = c_shape

        else:
            cl = [np.min(par_values), np.max(par_values)]
            norm = None
            cma = c_other

        cc = ax.pcolor(rr_pv, cmap=cma, vmin=cl[0], vmax=cl[1], norm=norm, edgecolor='k')
        fig.colorbar(cc, ax=ax)

        # add grid and title
        ax.set_title(par)
        ax.set_xticklabels([])
        ax.set_yticklabels([])
        ax.set_xticks([])
        ax.set_yticks([])

    ttl = 'GEV - {0}'.format(name)
    fig.suptitle(ttl, fontweight='bold', fontsize=14)

    # show
    if show: plt.show()
    return fig

def Plot_ChromosomesProbs(xds_chrom, show=True):
    'Plot chromosomes probabilities for each WT using chrom triangles'

    # TODO: colores poco fuertes
    # TODO: orden al colorcar los WTs ? mirar otras zonas de codigo
    # mirar tambien el orden de los WTs al generar chrom, si usamos sorted_bmus
    # desde un principio los WTs deberian respetar el orden de todo
    # TODO: lib/plotting/KMA.py Plot_KMArg_clusters_datamean para copiar estilo

    probs = xds_chrom.probs.values[:]
    chrom = xds_chrom.chrom.values[:]
    WTs = xds_chrom.WT
    ss = int(np.sqrt(len(WTs)))  # works for 36

    # triangle
    tsd = 10.0
    tsq = np.sqrt(tsd**2-(tsd/2)**2)
    tri_x = np.array([0, 0, tsq, 0])
    tri_y = np.array([0, tsd, tsd/2, 0])

    # chrom triangle positions (sea - swell_1 - swell_2)
    d_chrom_cntrs = {
        '100': (tsq, tsq/2),
        '010': (0, tsd),
        '001': (0, 0),
        '110': (tsq/2, tsd*3/4),
        '011': (0, tsd/2),
        '101': (tsq/2, tsd/4),
        '111': (tsq/3, tsd/2),
    }

    # max prob and size
    maxprob = np.nanmax(probs[:])
    maxsize = 1.5/maxprob

    # figure
    fig, axs = plt.subplots(ss, ss, figsize=(_faspect*_fsize*1.2, _fsize*1.2))
    axs = [i for sl in axs for i in sl]

    # plot each WT
    for c_wt, wt in enumerate(WTs):
        ax = axs[c_wt]

        # plot triangle
        ax.plot(tri_x, tri_y, 'k:', linewidth=1)

        # plot chromosomes prob circles
        for c_c, ch in enumerate(chrom):
            cntr = d_chrom_cntrs[''.join([str(int(x)) for x in ch])]
            if np.sum(ch) == 3:
                cl = [0,0,0]
            else:
                cl = ch
            ax.add_patch(
                plt.Circle(
                    cntr,
                    radius = maxsize * probs[c_wt, c_c],
                    edgecolor = cl,
                    facecolor = cl,
                    alpha = probs[c_wt, c_c] / maxprob,
                )
            )

        # waves family location text
        ax.text(tsq+0.75, tsq/2+0.5, 'SEA', va='center', ha='center', size=7)
        ax.text(0, tsd+0.5, 'SWELL_1', va='center', ha='center', size=7)
        ax.text(0, 0-0.5, 'SWELL_2', va='center', ha='center', size=7)

        # axis parameters
        ax.axis('equal')
        ax.axis('off')
        #ax.set_title('WT: {0}'.format(c_wt+1))

        ttl = 'Chromosomes Probabilities - WTs'
        fig.suptitle(ttl, fontweight='bold', fontsize=14)

    # show
    if show: plt.show()
    return fig

def Plot_SigmaCorrelation(xds_chrom, d_sigma, show=True):
    'Plot sigma correlation (Hs1-Hs2, Hs1-Tp1) for each WT using chrom triangles'

    # Get sigma correlation values for plot
    n_wts = len(d_sigma.keys())
    n_chs = len(d_sigma[list(d_sigma.keys())[0]].keys())
    ss = int(np.sqrt(n_wts))  # works for 36

    chrom = xds_chrom.chrom.values[:]

    sigma_plot = np.zeros((n_wts, n_chs)) * np.nan
    sigma_wtcrom = np.zeros((n_wts, n_chs)) * np.nan

    for c_wt, k_wt in enumerate(sorted(d_sigma.keys())):
        for c_ch, k_ch in enumerate(sorted(d_sigma[k_wt].keys())):

            corr = d_sigma[k_wt][k_ch]['corr']
            if corr.shape[0]==3:
                sigma_plot[c_wt, c_ch] = corr[0,1]  # one family. Hs - Tp corr
            elif corr.shape[0]==6:
                sigma_plot[c_wt, c_ch] = corr[0,3]  # two families. Hs1 - Hs2 corr

            sigma_wtcrom[c_wt, c_ch] = d_sigma[k_wt][k_ch]['wt_crom']

    # triangle
    tsd = 10.0
    tsq = np.sqrt(tsd**2-(tsd/2)**2)
    tri_x = np.array([0, 0, tsq, 0])
    tri_y = np.array([0, tsd, tsd/2, 0])

    # chrom triangle positions (sea - swell_1 - swell_2)
    d_chrom_cntrs = {
        '100': (tsq, tsq/2),
        '010': (0, tsd),
        '001': (0, 0),
        '110': (tsq/2, tsd*3/4),
        '011': (0, tsd/2),
        '101': (tsq/2, tsd/4),
        '111': (tsq/3, tsd/2),
    }
    # marker colors
    mk_pos = np.array([255, 0, 127]) / 255
    mk_neg = np.array([0, 127, 255]) / 255

    # max corr and size
    maxcorr = np.nanmax(sigma_plot[:])
    maxsize = 1.5/maxcorr

    # figure
    fig, axs = plt.subplots(ss, ss, figsize=(_faspect*_fsize*1.2, _fsize*1.2))
    axs = [i for sl in axs for i in sl]

    # plot each WT
    for c_wt, k_wt in enumerate(sorted(d_sigma.keys())):
        ax = axs[c_wt]

        # plot triangle
        ax.plot(tri_x, tri_y, 'k:', linewidth=1)

        # plot chromosomes prob circles
        for c_c, ch in enumerate(chrom):
            cntr = d_chrom_cntrs[''.join([str(int(x)) for x in ch])]

            # colors by sigma and wtcrom values
            if sigma_plot[c_wt, c_c] >= 0:
                mk_cl = mk_pos
            else:
                mk_cl = mk_neg

            if sigma_wtcrom[c_wt, c_c] == 0:
                e_cl = [0,0,0]
            else:
               e_cl = mk_cl

            # add circle    
            al = np.min([1.0, np.absolute(sigma_plot[c_wt, c_c]) / maxcorr])
            ax.add_patch(
                plt.Circle(
                    cntr,
                    radius = maxsize * np.absolute(sigma_plot[c_wt, c_c]),
                    edgecolor = e_cl,
                    facecolor = mk_cl,
                    alpha = al
                )
            )

        # waves family location text
        ax.text(tsq+0.75, tsq/2+0.5, 'SEA', va='center', ha='center', size=7)
        ax.text(0, tsd+0.5, 'SWELL_1', va='center', ha='center', size=7)
        ax.text(0, 0-0.5, 'SWELL_2', va='center', ha='center', size=7)
        # TODO: textos HH HT etc

        # axis parameters
        ax.axis('equal')
        ax.axis('off')
        #ax.set_title('WT: {0}'.format(c_wt+1))

        ttl = 'Sigma Correlation - WTs'
        fig.suptitle(ttl, fontweight='bold', fontsize=14)

    # show
    if show: plt.show()
    return fig

def Plot_Schemaball(xds_data):
    # TODO
    return None


# Climate Emulator: Simulaton Report

# TODO: refactor, upgrade, add to CE.Report_...

def axplot_compare_annualmax(ax, var_1, trp_1, var_2, trp_2, vn,
                             label_1='Historical', label_2='Simulation',
                             color_1='white', color_2='skyblue'):
    'axes plot histogram comparison between fit-sim variables'

    ax.semilogx(trp_1, var_1, 'ok', color = color_1, label = label_1,
                markersize = 4, zorder = 9)
    ax.semilogx(trp_2, var_2, '.-', color = color_2, label = label_2,
                linewidth = 2, zorder = 8)

    # customize axes
    ax.set_ylabel(vn)
    ax.grid(True)

def Plot_FitSim_AnnualMax(data_fit, data_sim, vns, vn_max=None,
                           color_1='white', color_2='skyblue',
                           label_1='Historical', label_2 = 'Simulation',
                           supt=False, show=True):
    'Plots fit vs sim annual maxima comparison for variables "vns"'

    # aux func for calculating rp time
    def t_rp(time_y):
        ny = len(time_y)
        return np.array([1/(1-(n/(ny+1))) for n in np.arange(1,ny+1)])

    # def. some auxiliar function to select all dataset variables at vn max by groups
    def grouped_max(ds, vn=None, dim=None):
        return ds.isel(**{dim: ds[vn].argmax(dim)})

    # grid spec number of rows
    gs_1 = len(vns)

    # plot figure
    fig = plt.figure(figsize=(_faspect*_fsize, _fsize*gs_1/2.3))

    # grid spec
    gs = gridspec.GridSpec(gs_1, 1)  #, wspace=0.0, hspace=0.0)

    # handle optional max variable and marginals
    if vn_max:
        amax_fit_marg = data_fit.groupby('time.year').apply(
            grouped_max, vn=vn_max, dim='time')
        amax_sim_marg = data_sim.groupby('time.year').apply(
            grouped_max, vn=vn_max, dim='time')

    # variables
    for c, vn in enumerate(vns):

        if vn_max:
            # get marginal variables at max
            amax_fit = amax_fit_marg[vn]
            amax_sim = amax_sim_marg[vn]

        else:
            # calculate Annual Maxima values 
            amax_fit = data_fit[vn].groupby('time.year').max(dim='time')
            amax_sim = data_sim[vn].groupby('time.year').max(dim='time')

        # get values and time, remove nans
        dh = amax_fit.values[:]; th = amax_fit['year'].values[:]
        th = th[~np.isnan(dh)]; dh = dh[~np.isnan(dh)]

        # remove last simulation year (1 time instant only to calculate max)
        ds = amax_sim.values[:-1]; ts = amax_sim['year'].values[:-1]
        ts = ts[~np.isnan(ds)]; ds = ds[~np.isnan(ds)]

        # prepare values for return period plot
        dh = np.sort(dh); th = t_rp(th)
        ds = np.sort(ds); ts = t_rp(ts)

        # axes
        ax = plt.subplot(gs[c, 0])
        axplot_compare_annualmax(
            ax, dh, th, ds, ts, vn,
            color_1=color_1, color_2=color_2,
            label_1=label_1, label_2=label_2,
        )

        # customize axes
        ax.legend(prop={'size':8})

    # last xaxis
    ax.set_xlabel('Return Period (Years)', fontsize=14)

    # fig suptitle
    if supt:
        fig.suptitle(
            '{0} - {1} Annual Max. Comparison: {2}'.format(label_1, label_2, ', '.join(vns)),
            fontsize=13, fontweight = 'bold',
        )

    # show and return figure
    if show: plt.show()
    return fig

def Plot_FitSim_GevFit(data_fit, data_sim, vn, xds_GEV_Par, kma_fit,
                       n_bins=30,
                       color_1='white', color_2='skyblue',
                       alpha_1=0.7, alpha_2=0.4,
                       label_1='Historical', label_2 = 'Simulation',
                       gs_1 = 1, gs_2 = 1, n_clusters = 1, vlim=1,
                       show=True):
    'Plots fit vs sim histograms and gev fit by clusters for variable "vn"'

    # plot figure
    fig = plt.figure(figsize=(_fsize*gs_2/2, _fsize*gs_1/2.3))

    # grid spec
    gs = gridspec.GridSpec(gs_1, gs_2)  #, wspace=0.0, hspace=0.0)

    # clusters
    for c in range(n_clusters):

        # select wt data
        wt = c+1

        ph_wt = np.where(kma_fit.bmus==wt)[0]
        ps_wt = np.where(data_sim.DWT==wt)[0]

        dh = data_fit[vn].values[:][ph_wt]  #; dh = dh[~np.isnan(dh)]
        ds = data_sim[vn].values[:][ps_wt] #; ds = ds[~np.isnan(ds)]

        # TODO: problem if gumbell?
        # select wt GEV parameters
        pars_GEV = xds_GEV_Par[vn]
        sha = pars_GEV.sel(parameter='shape').sel(n_cluster=wt).values
        sca = pars_GEV.sel(parameter='scale').sel(n_cluster=wt).values
        loc = pars_GEV.sel(parameter='location').sel(n_cluster=wt).values


        # compare histograms
        ax = fig.add_subplot(gs[c])
        axplot_compare_histograms(
            ax, dh, ds, ttl='WT: {0}'.format(wt), density=True, n_bins=n_bins,
            color_1=color_1, color_2=color_2,
            alpha_1=alpha_1, alpha_2=alpha_2,
            label_1=label_1, label_2=label_2,
        )

        # add gev fit 
        x = np.linspace(genextreme.ppf(0.001, -1*sha, loc, sca), vlim, 100)
        ax.plot(x, genextreme.pdf(x, -1*sha, loc, sca), label='GEV fit')

        # customize axis
        ax.legend(prop={'size':8})

    # fig suptitle
    #fig.suptitle('{0}'.format(vn), fontsize=14, fontweight = 'bold')

    # show and return figure
    if show: plt.show()
    return fig


def Plot_Fit_QQ(data_fit, vn, xds_GEV_Par, kma_fit, color='black',
                gs_1 = 1, gs_2 = 1, n_clusters = 1,
                show=True):
    'Plots QQ (empirical-gev) for variable vn and each kma cluster'

    # plot figure
    fig = plt.figure(figsize=(_fsize*gs_2/2, _fsize*gs_1/2.3))

    # grid spec
    gs = gridspec.GridSpec(gs_1, gs_2)  #, wspace=0.0, hspace=0.0)

    # clusters
    for c in range(n_clusters):

        # select wt data
        wt = c+1
        ph_wt = np.where(kma_fit.bmus==wt)[0]
        dh = data_fit[vn].values[:][ph_wt]; dh = dh[~np.isnan(dh)]

        # prepare data
        Q_emp = np.sort(dh)
        bs = np.linspace(1, len(dh), len(dh))
        pp = bs / (len(dh)+1)

        # TODO: problem if gumbell?
        # select wt GEV parameters
        pars_GEV = xds_GEV_Par[vn]
        sha = pars_GEV.sel(parameter='shape').sel(n_cluster=wt).values
        sca = pars_GEV.sel(parameter='scale').sel(n_cluster=wt).values
        loc = pars_GEV.sel(parameter='location').sel(n_cluster=wt).values

        # calc GEV pdf
        Q_gev = genextreme.ppf(pp, -1*sha, loc, sca)

        # scatter plot
        ax = fig.add_subplot(gs[c])
        ax.plot(Q_emp, Q_gev, 'ok', color = color, label='N = {0}'.format(len(dh)))
        ax.plot([0, 1], [0, 1], '--b', transform=ax.transAxes)

        # customize axis
        ax.set_title('WT: {0}'.format(wt))
        ax.axis('equal')
        #ax.set_xlabel('Empirical')
        ax.set_ylabel('GEV')
        ax.legend(prop={'size':8})

    # fig suptitle
    #fig.suptitle('{0}'.format(vn), fontsize=14, fontweight = 'bold')

    # show and return figure
    if show: plt.show()
    return fig


# Extremes Return Period for all simulations

def axplot_RP(ax, t_h, v_h, tg_h, vg_h, t_s, v_s, var_name, sim_percentile=95):
    'axes plot return period historical vs simulation'

    # historical maxima
    ax.semilogx(
        t_h, v_h, 'ok',
        markersize = 3, label = 'Historical',
        zorder=9,
    )

    # TODO: fit historical to gev
    # historical GEV fit
    #ax.semilogx(
    #    tg_h, vg_h, '-b',
    #    label = 'Historical - GEV Fit',
    #)

    # simulation maxima - mean
    mn = np.mean(v_s, axis=0)
    ax.semilogx(
        t_s, mn, '-r',
        linewidth = 2, label = 'Simulation (mean)',
        zorder=8,
    )

    # simulation maxima percentiles
    out = 100 - sim_percentile
    p95 = np.percentile(v_s, 100-out/2.0, axis=0,)
    p05 = np.percentile(v_s, out/2.0, axis=0,)

    ax.semilogx(
        t_s, p95, linestyle='-', color='grey',
        linewidth = 2, #label = 'Simulation (95% percentile)',
    )

    ax.semilogx(
        t_s, p05, linestyle='-', color='grey',
        linewidth = 2, # label = 'Simulation (05% percentile)',
    )
    ax.fill_between(
        t_s, p05, p95, color='lightgray',
        label = 'Simulation ({0}% C.I)'.format(sim_percentile)
    )

    # customize axs
    ax.legend(loc='lower right')
    ax.set_title('Annual Maxima', fontweight='bold')
    ax.set_xlabel('Return Period (years)')
    ax.set_ylabel('{0}'.format(var_name))
    ax.set_xlim(left=10**0, right=np.max(np.concatenate([t_h,t_s])))
    ax.tick_params(axis='both', which='both', top=True, right=True)
    ax.grid(which='both')

def Plot_ReturnPeriodValidation(xds_hist, xds_sim, sim_percentile=95, show=True):
    'Plot Return Period historical - simulation validation'

    # aux func for calculating rp time
    def t_rp(time_y):
        ny = len(time_y)
        return np.array([1/(1-(n/(ny+1))) for n in np.arange(1,ny+1)])

    # aux func for gev fit
    # TODO: fix it
    def gev_fit(var_fit):
        c = -0.1
        vv = np.linspace(0,10,200)

        sha_g, loc_g, sca_g =  genextreme.fit(var_fit, c)
        pg = genextreme.cdf(vv, sha_g, loc_g, sca_g)

        ix = pg > 0.1
        vv = vv[ix]
        ts = 1/(1 - pg[ix])

        # TODO gev params 95% confidence intervals

        return ts, vv

    # clean nans
    t_r = xds_hist.year.values[:]
    v_r = xds_hist.values[:]

    ix_nan = np.isnan(v_r)
    t_r = t_r[~ix_nan]
    v_r = v_r[~ix_nan]

    # RP calculation, var sorting historical
    t_h = t_rp(t_r)
    v_h = np.sort(v_r)

    # GEV fit historical
    #tg_h, vg_h = gev_fit(v_h)
    tg_h, vg_h = [],[]

    # RP calculation, var sorting simulation
    t_s = t_rp(xds_sim.year.values[:-1])  # remove last year*
    v_s = np.sort(xds_sim.values[:,:-1])  # remove last year*

    # figure
    fig, axs = plt.subplots(figsize=(_faspect*_fsize, _fsize))

    axplot_RP(
        axs,
        t_h, v_h, tg_h, vg_h,
        t_s, v_s,
        xds_sim.name,
        sim_percentile=sim_percentile,
    )

    # show and return figure
    if show: plt.show()
    return fig








# TODO: revisar funciones _simulations


def axplot_RP_Sims(ax, tups_time_val, var_name):
    'axes plot return period historical vs simulation'


    lbls_cls = [
        #('Simulation', 'black'),
        ('Simulation - El Niño', 'red'),
        ('Simulation - La Niña' , 'blue'),
    ]

    for (t_s, v_s), (lb, cl) in zip(tups_time_val, lbls_cls):


        # simulation maxima - mean
        mn = np.mean(v_s, axis=0)
        ax.semilogx(
            t_s, mn, linestyle= '-', color=cl,
            linewidth = 2, label = lb,
            zorder=8,
        )

        # simulation maxima percentile 95% and 05%
        p95 = np.percentile(v_s, 95, axis=0,)
        p05 = np.percentile(v_s, 5, axis=0,)

        ax.semilogx(
            t_s, p95, linestyle='--', color=cl, alpha=0.05,
            linewidth = 2, #label = 'Simulation (95% percentile)',
        )

        ax.semilogx(
            t_s, p05, linestyle='--', color=cl, alpha=0.05,
            linewidth = 2, # label = 'Simulation (05% percentile)',
        )
        ax.fill_between(
            t_s, p05, p95, color=cl,alpha=0.05,
            #label = 'Simulation (05 - 95 percentile)'
        )

    # customize axs
    ax.legend(loc='lower right')
    ax.set_title('Annual Maxima', fontweight='bold')
    ax.set_xlabel('Return Period (years)')
    ax.set_ylabel('{0}'.format(var_name))
    ax.set_xlim(left=10**0, right=np.max(t_s))
    ax.tick_params(axis='both', which='both', top=True, right=True)
    ax.grid(which='both')

def Plot_ReturnPeriod_Simulations(l_xds_sim, show=True):
    'Plot Return Period historical - simulation validation'

    # aux func for calculating rp time
    def t_rp(time_y):
        ny = len(time_y)
        return np.array([1/(1-(n/(ny+1))) for n in np.arange(1,ny+1)])

    # RP calculation, var sorting simulation
    tups_time_val = []
    for xds_sim in l_xds_sim:
        t_s = t_rp(xds_sim.year.values[:-1])  # remove last year*
        v_s = np.sort(xds_sim.values[:,:-1])  # remove last year*

        tups_time_val.append((t_s, v_s))

    # figure
    fig, axs = plt.subplots(figsize=(_faspect*_fsize, _fsize))

    axplot_RP_Sims(
        axs,
        tups_time_val,
        xds_sim.name,
    )

    # show and return figure
    if show: plt.show()
    return fig



# TODO: revisar funciones _cambio climatico


def axplot_RP_CC(ax, t_h, v_h, tg_h, vg_h, t_s, v_s, t_s2, v_s2, var_name, sim_percentile=95,label_1='Simulation', label_2 = 'Simulation Climate Change', ):
    'axes plot return period historical vs simulation'

    # historical maxima
    ax.semilogx(
        t_h, v_h, 'ok',
        markersize = 3, label = 'Historical',
        zorder=9,
    )

    # TODO: fit historical to gev
    # historical GEV fit
    #ax.semilogx(
    #    tg_h, vg_h, '-b',
    #    label = 'Historical - GEV Fit',
    #)

    # simulation maxima - mean
    mn = np.mean(v_s, axis=0)
    ax.semilogx(
        t_s, mn, '-r',
        linewidth = 2, label = '{} (mean)'.format(label_1),
        zorder=8,
    )

    # simulation climate change maxima - mean
    mn2 = np.mean(v_s2, axis=0)
    ax.semilogx(
        t_s2, mn2, '-b',
        linewidth = 2, label = '{} (mean)'.format(label_2),
        zorder=8,
    )

    # simulation maxima percentiles
    out = 100 - sim_percentile
    p95 = np.percentile(v_s, 100-out/2.0, axis=0,)
    p05 = np.percentile(v_s, out/2.0, axis=0,)

    ax.semilogx(
        t_s, p95, linestyle='-', color='grey',
        linewidth = 2, #label = 'Simulation (95% percentile)',
    )

    ax.semilogx(
        t_s, p05, linestyle='-', color='grey',
        linewidth = 2, # label = 'Simulation (05% percentile)',
    )
    ax.fill_between(
        t_s, p05, p95, color='lightgray',
        #label = 'Simulation ({0}% C.I)'.format(sim_percentile)
        label = '{} ({} C.I)'.format(label_1, sim_percentile)
    )

    # simulation climate change maxima percentiles
    out = 100 - sim_percentile
    p95 = np.percentile(v_s2, 100-out/2.0, axis=0,)
    p05 = np.percentile(v_s2, out/2.0, axis=0,)

    ax.semilogx(
        t_s2, p95, linestyle='-', color='skyblue',
        linewidth = 2, #label = 'Simulation (95% percentile)',
    )

    ax.semilogx(
        t_s2, p05, linestyle='-', color='skyblue',
        linewidth = 2, # label = 'Simulation (05% percentile)',
    )
    ax.fill_between(
        t_s2, p05, p95, color='skyblue', alpha=0.5,
        #label = 'Simulation Climate Change ({0}% C.I)'.format(sim_percentile)
        label = '{} ({} C.I)'.format(label_2, sim_percentile)
    )

    # customize axs
    ax.legend(loc='lower right')
    ax.set_title('Annual Maxima', fontweight='bold')
    ax.set_xlabel('Return Period (years)')
    ax.set_ylabel('{0}'.format(var_name))
    ax.set_xlim(left=10**0, right=np.max(np.concatenate([t_h,t_s])))
    ax.tick_params(axis='both', which='both', top=True, right=True)
    ax.grid(which='both')


def Plot_ReturnPeriodValidation_CC(xds_hist, xds_sim, xds_sim2, sim_percentile=95, label_1='Simulation', label_2 = 'Simulation Climate Change', show=True):
    'Plot Return Period historical - simulation validation - simulation CLIMATE CHANGE'

    # aux func for calculating rp time
    def t_rp(time_y):
        ny = len(time_y)
        return np.array([1/(1-(n/(ny+1))) for n in np.arange(1,ny+1)])

    # aux func for gev fit
    # TODO: fix it
    def gev_fit(var_fit):
        c = -0.1
        vv = np.linspace(0,10,200)

        sha_g, loc_g, sca_g = genextreme.fit(var_fit, c)
        pg = genextreme.cdf(vv, sha_g, loc_g, sca_g)

        ix = pg > 0.1
        vv = vv[ix]
        ts = 1/(1 - pg[ix])

        # TODO gev params 95% confidence intervals

        return ts, vv

    # clean nans
    t_r = xds_hist.year.values[:]
    v_r = xds_hist.values[:]

    ix_nan = np.isnan(v_r)
    t_r = t_r[~ix_nan]
    v_r = v_r[~ix_nan]

    # RP calculation, var sorting historical
    t_h = t_rp(t_r)
    v_h = np.sort(v_r)

    # GEV fit historical
    #tg_h, vg_h = gev_fit(v_h)
    tg_h, vg_h = [],[]

    # RP calculation, var sorting simulation
    t_s = t_rp(xds_sim.year.values[:-1])  # remove last year*
    v_s = np.sort(xds_sim.values[:,:-1])  # remove last year*

    t_s2 = t_rp(xds_sim2.year.values[:-1])  # remove last year*
    v_s2 = np.sort(xds_sim2.values[:,:-1])  # remove last year*

    # figure
    fig, axs = plt.subplots(figsize=(_faspect*_fsize, _fsize))

    axplot_RP_CC(
        axs,
        t_h, v_h, tg_h, vg_h,
        t_s, v_s,
        t_s2, v_s2,
        xds_sim.name,
        sim_percentile=sim_percentile,
        label_1 = label_1,
        label_2 = label_2,
    )

    # show and return figure
    if show: plt.show()
    return fig



import pickle

from .models import BlueMathModel


def load_model(model_path: str) -> BlueMathModel:
    """
    Loads a BlueMathModel from a file.

    Parameters
    ----------
    model_path : str
        The path to the model file.

    Returns
    -------
    BlueMathModel
        The loaded BlueMathModel.
    """

    with open(model_path, "rb") as f:
        model = pickle.load(f)

    return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# common
import os
import os.path as op
from datetime import datetime, timedelta

# pip
import numpy as np

import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import matplotlib.dates as mdates
import matplotlib.colors as mcolors
from matplotlib.colorbar import ColorbarBase
from matplotlib.ticker import MaxNLocator

# teslakit
from ..util.operations import GetBestRowsCols
from .custom_colors import GetClusterColors
from ..kma import ClusterProbabilities, ChangeProbabilities
#from .outputs import axplot_compare_histograms


# import constants
from .config import _faspect, _fsize, _fdpi


def GenOneYearDaily(yy=1981, month_ini=1):
    'returns one generic year in a list of datetimes. Daily resolution'

    dp1 = datetime(yy, month_ini, 1)
    dp2 = dp1+timedelta(days=365)

    return [dp1 + timedelta(days=i) for i in range((dp2-dp1).days)]

def Generate_PerpYear_Matrix(num_clusters, bmus_values, bmus_dates,
                             num_sim=1, month_ini=1):
    '''
    Calculates and returns matrix for stacked bar plotting

    bmus_dates - datetime.datetime (only works if daily resolution)
    bmus_values has to be 2D (time, nsim)
    '''

    # generate perpetual year list
    list_pyear = GenOneYearDaily(month_ini=month_ini)

    # generate aux arrays
    m_plot = np.zeros((num_clusters, len(list_pyear))) * np.nan
    bmus_dates_months = np.array([d.month for d in bmus_dates])
    bmus_dates_days = np.array([d.day for d in bmus_dates])

    # sort data
    for i, dpy in enumerate(list_pyear):
        _, s = np.where(
            [(bmus_dates_months == dpy.month) & (bmus_dates_days == dpy.day)]
        )
        b = bmus_values[s,:]
        b = b.flatten()

        for j in range(num_clusters):
            _, bb = np.where([(j+1 == b)])  # j+1 starts at 1 bmus value!

            m_plot[j,i] = float(len(bb)/float(num_sim))/len(s)

    return m_plot


def axplot_PerpYear(ax, num_clusters, bmus_values, bmus_dates, num_sim, month_ini):
    'axes plot bmus perpetual year'

    # get cluster colors for stacked bar plot
    np_colors_int = GetClusterColors(num_clusters)

    # generate dateticks
    x_val = GenOneYearDaily(month_ini = month_ini)

    # generate plot matrix
    m_plot = Generate_PerpYear_Matrix(
        num_clusters, bmus_values, bmus_dates,
        num_sim = num_sim, month_ini = month_ini)

    # plot stacked bars
    bottom_val = np.zeros(m_plot[1,:].shape)
    for r in range(num_clusters):
        row_val = m_plot[r,:]
        ax.bar(
            x_val, row_val, bottom = bottom_val,
            width = 1, color = np.array([np_colors_int[r]])
        )

        # store bottom
        bottom_val += row_val

    # customize  axis
    months = mdates.MonthLocator()
    monthsFmt = mdates.DateFormatter('%b')

    ax.set_xlim(x_val[0], x_val[-1])
    ax.xaxis.set_major_locator(months)
    ax.xaxis.set_major_formatter(monthsFmt)
    ax.set_ylim(0, 1)
    ax.set_ylabel('')

def axplot_ChangeProbs(ax, change_probs, wt_colors,
                       ttl = '', vmin = 0, vmax = 1,
                       cmap = 'Blues', cbar_ttl = ''):
    'axes plot cluster change probabilities'

    num_clusters = change_probs.shape[0]

    # clsuter transition plot
    pc = ax.pcolor(
        change_probs,
        cmap=cmap, vmin=vmin, vmax=vmax,
    )

    # add colorbar
    cbar = plt.colorbar(pc, ax=ax)
    cbar.ax.tick_params(labelsize=8)
    if vmin != 0 or vmax !=1:
        cbar.set_ticks(np.linspace(vmin, vmax, 6))
    cbar.ax.set_ylabel(cbar_ttl, rotation=270, labelpad=20)

    # customize axes
    ax.set_xticks(np.arange(num_clusters)+0.5)
    ax.set_yticks(np.arange(num_clusters)+0.5)
    ax.set_xticklabels([])
    ax.set_yticklabels([])
    ax.set_title(ttl, {'fontsize':10, 'fontweight':'bold'})

    # add custom color axis
    ccmap = mcolors.ListedColormap([tuple(r) for r in wt_colors])

    # custom color axis positions
    ax_pos = ax.get_position()
    cax_x_pos = [ax_pos.x0, ax_pos.y0-0.03, ax_pos.width, 0.02]
    cax_y_pos = [ax_pos.x0-0.03/_faspect, ax_pos.y0, 0.02/_faspect, ax_pos.height]

    # custom color axis X
    cax_x = ax.figure.add_axes(cax_x_pos)
    cbar_x = ColorbarBase(
        cax_x, cmap = ccmap, orientation='horizontal',
        norm = mcolors.Normalize(vmin=0, vmax=num_clusters),
    )
    cbar_x.set_ticks([])

    # custom color axis Y
    cax_y= ax.figure.add_axes(cax_y_pos)
    cbar_y = ColorbarBase(
        cax_y, cmap = ccmap, orientation='vertical',
        norm = mcolors.Normalize(vmin=0, vmax=num_clusters),
    )
    cbar_y.set_ticks([])

def axplot_Transition(ax, probs_change_fit, probs_change_sim,
                      ttl='', color='black'):
    'axes plot for scatter historical - simulation transition matrix comparison'

    # clsuter transition fit vs. sim scatter plot
    pc = ax.scatter(
        probs_change_fit, probs_change_sim,
        color=color, s=5, zorder=2,
    )

    # get max value for axis lim
    axlim = np.ceil(10*np.max(probs_change_sim[:]))/10

    # customize axes
    ax.plot(([0,0],[1,1]), linestyle='dashed', linewidth=0.5, color='grey',
            zorder=1)
    ax.set_ylabel('Simulated', {'fontsize':8})
    ax.set_xlabel('Historical', {'fontsize':8})
    ax.grid(True, which='major', axis='both', linestyle='--', color='grey')
    ax.tick_params(axis='both', which='major', labelsize=7)
    ax.set_xlim([0, axlim])
    ax.set_ylim([0, axlim])
    ax.set_title(ttl, {'fontsize':10, 'fontweight':'bold'})

def axplot_ClusterProbs(ax, cluster_probs_fit, cluster_probs_sim, wt_colors,
                       ttl=''):
    'axes plot for scatter historical - simulation cluster probs comparison'

    # clsuter transition plot
    for cf, cs, wc in zip(cluster_probs_fit, cluster_probs_sim, wt_colors):
        ax.plot(cf, cs, marker='.', markersize=8, color=wc, zorder=2)

    # get max value for axis lim
    axlim = np.ceil(10*np.max(cluster_probs_sim[:]))/10

    # customize axes
    ax.plot(([0,0],[1,1]), linestyle='dashed', linewidth=0.5, color='grey',
            zorder=1)
    ax.set_ylabel('Simulated', {'fontsize':8})
    ax.set_xlabel('Historical', {'fontsize':8})
    ax.grid(True, which='major', axis='both', linestyle='--', color='grey')
    ax.tick_params(axis='both', which='major', labelsize=7)
    ax.set_xlim([0, axlim])
    ax.set_ylim([0, axlim])
    ax.set_title(ttl, {'fontsize':10, 'fontweight':'bold'})

def axplot_WT_Probs(ax, wt_probs,
                     ttl = '', vmin = 0, vmax = 0.1,
                     cmap = 'Blues', caxis='black'):
    'axes plot WT cluster probabilities'

    # clsuter transition plot
    pc = ax.pcolor(
        np.flipud(wt_probs),
        cmap=cmap, vmin=vmin, vmax=vmax,
        edgecolors='k',
    )

    # customize axes
    ax.set_xticks([])
    ax.set_yticks([])
    ax.set_title(ttl, {'fontsize':10, 'fontweight':'bold'})

    # axis color
    plt.setp(ax.spines.values(), color=caxis)
    plt.setp(
        [ax.get_xticklines(), ax.get_yticklines()],
        color=caxis,
    )

    # axis linewidth
    if caxis != 'black':
        plt.setp(ax.spines.values(), linewidth=3)

    return pc

def axplot_WT_Hist(ax, bmus, n_clusters, ttl=''):
    'axes plot WT cluster count histogram'

    # cluster transition plot
    ax.hist(
        bmus,
        bins = np.arange(1, n_clusters+2),
        edgecolor='k'
    )

    # customize axes
    #ax.grid('y')

    ax.set_xticks(np.arange(1,n_clusters+1)+0.5)
    ax.set_xticklabels(np.arange(1,n_clusters+1))
    ax.set_xlim([1, n_clusters+1])
    ax.tick_params(axis='both', which='major', labelsize=6)

    ax.set_title(ttl, {'fontsize':10, 'fontweight':'bold'})


def Plot_Compare_PerpYear(num_clusters,
                          bmus_values_sim, bmus_dates_sim,
                          bmus_values_hist, bmus_dates_hist,
                          n_sim = 1, month_ini = 1, show=True):
    '''
    Plot simulated - historical bmus comparison in a perpetual year

    bmus_dates requires 1 day resolution time
    bmus_values set min value has to be 1 (not 0)
    '''

    # check dates have 1 day time resolution 
    td_h = bmus_dates_hist[1] - bmus_dates_hist[0]
    td_s = bmus_dates_sim[1] - bmus_dates_sim[0]
    
    if isinstance(td_h, np.int32):
        return
    if td_h.days != 1 or td_s.days != 1:
        print('PerpetualYear bmus comparison skipped.')
        print('timedelta (days): Hist - {0}, Sim - {1})'.format(
            td_h, td_s.days))
        return

    # plot figure
    fig, (ax_hist, ax_sim) = plt.subplots(2,1, figsize=(_faspect*_fsize, _fsize))

    # historical perpetual year
    axplot_PerpYear(
        ax_hist, num_clusters, bmus_values_hist, bmus_dates_hist,
        num_sim = 1, month_ini = month_ini,
    )
    ax_hist.set_title('Historical')

    # simulated perpetual year
    axplot_PerpYear(
        ax_sim, num_clusters, bmus_values_sim, bmus_dates_sim,
        num_sim = n_sim, month_ini = month_ini,
    )
    ax_sim.set_title('Simulation')

    # add custom colorbar
    np_colors_int = GetClusterColors(num_clusters)
    ccmap = mcolors.ListedColormap(
        [tuple(r) for r in np_colors_int]
    )
    cax = fig.add_axes([0.92, 0.125, 0.025, 0.755])
    cbar = ColorbarBase(
        cax, cmap=ccmap,
        norm = mcolors.Normalize(vmin=0, vmax=num_clusters),
        ticks = np.arange(num_clusters) + 0.5,
    )
    cbar.ax.tick_params(labelsize=8)
    cbar.set_ticklabels(range(1, num_clusters + 1))

    # text
    fig.suptitle('Perpetual Year', fontweight='bold', fontsize=12)

    # show and return figure
    if show: plt.show()
    return fig

def Plot_Compare_Transitions(num_clusters, bmus_values_hist, bmus_values_sim,
                             sttl=None, show=True):
    '''
    Plot many probabilities transition comparisons between 2 bmus series

    bmus_values_hist, bmus_values_sim (series x n_sim) - bmus series to compare
    num_clusters  - total number of clusters at series
    '''

    # if n_sim > 1 lets safe-join simulated bmus for probs calcs.
    n_sim = bmus_values_sim.shape[1]
    if n_sim > 1:
        # add nan splitter and then flatten
        bmus_values_sim = np.append(
            bmus_values_sim, np.zeros((1, n_sim))*np.nan,
            axis = 0,
        )
        bmus_values_sim = bmus_values_sim.flatten('F')

    # get cluster colors
    wt_colors = GetClusterColors(num_clusters)
    wt_set = np.arange(num_clusters) + 1

    # cluster probabilities
    probs_c_fit = ClusterProbabilities(bmus_values_hist, wt_set)
    probs_c_sim = ClusterProbabilities(bmus_values_sim, wt_set)

    # cluster change probabilities
    _, chprobs_fit = ChangeProbabilities(bmus_values_hist, wt_set)
    _, chprobs_rnd = ChangeProbabilities(bmus_values_sim, wt_set)
    chprobs_dif = chprobs_rnd - chprobs_fit

    # figure
    fig = plt.figure(figsize=(_faspect*_fsize, _fsize))

    # layout
    gs = gridspec.GridSpec(2, 6, wspace=0.60, hspace=0.35)
    ax_chpr_fit = plt.subplot(gs[0, :2])
    ax_chpr_rnd = plt.subplot(gs[0, 2:4])
    ax_chpr_dif = plt.subplot(gs[0, 4:])
    ax_scat_tra = plt.subplot(gs[1, 1:3])
    ax_scat_pbs = plt.subplot(gs[1, 3:5])

    # cluster change probs axes 
    axplot_ChangeProbs(ax_chpr_fit, chprobs_fit, wt_colors,
                       'Historical WT Transition Probabilities')
    axplot_ChangeProbs(ax_chpr_rnd, chprobs_rnd, wt_colors,
                       'Simulation WT Transition Probabilities')
    axplot_ChangeProbs(ax_chpr_dif, chprobs_dif, wt_colors,
                       'Change in WT Transition Probabilities',
                       vmin=-0.05, vmax=0.05, cmap='RdBu_r',
                       cbar_ttl='Simulation - Historical',
                      )

    # scatter plot transitions
    axplot_Transition(ax_scat_tra, chprobs_fit, chprobs_rnd,
                      'Sim. vs. Hist. WT Transition Probabilities'
                     )

    # scatter plot cluster probs
    axplot_ClusterProbs(ax_scat_pbs, probs_c_fit, probs_c_sim, wt_colors,
                       'Sim. vs. Hist. WT Probabilities',
                       )

    # suptitle
    if sttl != None:
        fig.suptitle(sttl, fontweight='bold', fontsize=12)

    # show and return figure
    if show: plt.show()
    return fig

def Plot_Probs_WT_WT(series_1, series_2, n_clusters_1, n_clusters_2, ttl='',
                     wt_colors=False, show=True):
    '''
    Plot WTs_1 / WTs_2 probabilities

    both categories series should start at 0
    '''

    # set of daily weather types
    set_2 = np.arange(n_clusters_2)

    # dailt weather types matrix rows and cols
    n_rows, n_cols = GetBestRowsCols(n_clusters_2)

    # get cluster colors
    cs_wt =  GetClusterColors(n_clusters_1)

    # plot figure
    fig = plt.figure(figsize=(_faspect*_fsize, _fsize/3))
    gs = gridspec.GridSpec(1, n_clusters_1, wspace=0.10, hspace=0.15)

    for ic in range(n_clusters_1):

        # select DWT bmus at current AWT indexes
        index_1 = np.where(series_1==ic)[0][:]
        sel_2 = series_2[index_1]

        # get DWT cluster probabilities
        cps = ClusterProbabilities(sel_2, set_2)
        C_T = np.reshape(cps, (n_rows, n_cols))

        # axis colors
        if wt_colors:
            caxis = cs_wt[ic]
        else:
            caxis = 'black'

        # plot axes
        ax = plt.subplot(gs[0, ic])
        axplot_WT_Probs(
            ax, C_T,
            ttl = 'WT {0}'.format(ic+1),
            cmap = 'Reds', caxis = caxis,
        )
        ax.set_aspect('equal')

    # add fig title
    fig.suptitle(ttl, fontsize=14, fontweight='bold')

    # show and return figure
    if show: plt.show()
    return fig

def Plot_Probs_WT_WT_anomaly(series_1, series_2, n_clusters_1, n_clusters_2, ttl='',
                     wt_colors=False, show=True):
    '''
    Plot WTs_1 / WTs_2 probabilities

    both categories series should start at 0
    '''

    # set of daily weather types
    set_2 = np.arange(n_clusters_2)

    # dailt weather types matrix rows and cols
    n_rows, n_cols = GetBestRowsCols(n_clusters_2)

    # get cluster colors
    cs_wt =  GetClusterColors(n_clusters_1)

    # plot figure
    fig = plt.figure(figsize=(_faspect*_fsize, _fsize/3))
    gs = gridspec.GridSpec(1, n_clusters_1, wspace=0.10, hspace=0.15)

    #----------------------
    # best rows cols combination
    n_rows, n_cols = GetBestRowsCols(n_clusters_2)

    c_T = ClusterProbabilities(series_2, set_2)
    C_T_mean = np.reshape(c_T, (n_rows, n_cols))
    #----------------------

    for ic in range(n_clusters_1):

        # select DWT bmus at current AWT indexes
        index_1 = np.where(series_1==ic)[0][:]
        sel_2 = series_2[index_1]

        # get DWT cluster probabilities
        cps = ClusterProbabilities(sel_2, set_2)
        C_T = np.reshape(cps, (n_rows, n_cols))

        #-----------
        C_T = C_T-C_T_mean
        #-----------

        # axis colors
        if wt_colors:
            caxis = cs_wt[ic]
        else:
            caxis = 'black'

        # plot axes
        ax = plt.subplot(gs[0, ic])
        axplot_WT_Probs(
            ax, C_T,
            ttl = 'WT {0}'.format(ic+1),vmin=-0.1,vmax=0.1,
            cmap = 'seismic', caxis = caxis,
        )
        ax.set_aspect('equal')

    # add fig title
    fig.suptitle(ttl, fontsize=14, fontweight='bold')

    # show and return figure
    if show: plt.show()
    return fig

def Plot_Compare_Persistences(n_clusters, pers_hist, pers_sim,
                              color_1='plum', color_2='dodgerblue',
                              alpha_1=0.7, alpha_2=0.4,
                              label_1='Historical', label_2 = 'Simulation',
                              show=True):
    '''
    Plot historical - simulation WT persistences comparison

    pers_sim, pers_hist - dictionary {WT: ocurrences durations list}
    '''

    # dailt weather types matrix rows and cols
    n_rows, n_cols = GetBestRowsCols(n_clusters)

    # plot figure
    fig = plt.figure(figsize=(_fsize*n_cols/2, _fsize*n_rows/2.3))
    gs = gridspec.GridSpec(n_rows, n_cols, wspace=0.2, hspace=0.2)

    for c in range(n_clusters):

        # select wt data
        wt = c+1

        # persistences
        p_h = pers_hist[wt]
        p_s = pers_sim[wt]

        # common set
        ss = np.array(sorted(list(set(p_h).union(set(p_s)))))+0.5
        # filter when only 1 bin 
        if len(ss) == 1: continue

        # compare histograms
        ax = fig.add_subplot(gs[c])
        axplot_compare_histograms(
            ax, sorted(p_h), sorted(p_s), ttl='WT: {0}'.format(wt),
            density=True,
            n_bins=ss,
            color_1=color_1, color_2=color_2,
            alpha_1=alpha_1, alpha_2=alpha_2,
            label_1=label_1, label_2=label_2,
        )

        # customize axis
        ax.set_ylabel('')
        ax.legend(prop={'size':8})
        ax.xaxis.set_major_locator(MaxNLocator(integer=True))

    # suptitle
    fig.suptitle('WTs Persistences (days)', fontweight='bold', fontsize=12)

    # show and return figure
    if show: plt.show()
    return fig


# TODO: following functions are not finished / tested

def Plot_PerpYear(bmus_values, bmus_dates, num_clusters, num_sim=1,
                  show=True):
    'Plots ARL bmus simulated in a perpetual_year stacked bar chart'

    # TODO: UPDATE


    # get cluster colors for stacked bar plot
    np_colors_int = GetClusterColors(num_clusters)

    # generate plot matrix
    m_plot = Generate_PerpYear_Matrix(
        num_clusters, bmus_values, bmus_dates, num_sim)

    # plot figure
    fig, ax = plt.subplots(1,1, figsize=(_faspect*_fsize, _fsize))

    bottom_val = np.zeros(m_plot[1,:].shape)
    for r in range(num_clusters):
        row_val = m_plot[r,:]
        plt.bar(
            range(365), row_val, bottom=bottom_val,
            width=1, color = np.array([np_colors_int[r]])
               )

        # store bottom
        bottom_val += row_val

    # axis
    # TODO: REFINAR AXIS
    plt.xlim(1, 365)
    plt.ylim(0, 1)
    plt.xlabel('Perpetual year')
    plt.ylabel('')

    # show and return figure
    if show: plt.show()
    return fig

